"""A pipeline consists of [`Node`][amltk.pipeline.node.Node]s, which hold
the various attributes required to build a pipeline, such as the
[`.item`][amltk.pipeline.node.Node.item], its [`.space`][amltk.pipeline.node.Node.space],
its [`.config`][amltk.pipeline.node.Node.config] and so on.

The [`Node`][amltk.pipeline.node.Node]s are connected to each in a parent-child
relation ship where the children are simply the [`.nodes`][amltk.pipeline.node.Node.nodes]
that the parent leads to.

To give these attributes and relations meaning, there are various subclasses
of [`Node`][amltk.pipeline.node.Node] which give different syntactic meanings
when you want to construct something like a
[`search_space()`][amltk.pipeline.node.Node.search_space] or
[`build()`][amltk.pipeline.node.Node.build] some concrete object out of the
pipeline.

For example, a [`Sequential`][amltk.pipeline.Sequential] node
gives the meaning that each of its children in
[`.nodes`][amltk.pipeline.node.Node.nodes] should follow one another while
something like a [`Choice`][amltk.pipeline.Choice]
gives the meaning that only one of its children should be chosen.

You will likely never have to create a [`Node`][amltk.pipeline.node.Node]
directly, but instead use the various components to create the pipeline.

??? note "Hashing"

    When hashing a node, i.e. to put it in a `set` or as a key in a `dict`,
    only the name of the node and the hash of its children is used.
    This means that two nodes with the same connectivity will be equalling hashed,

??? note "Equality"

    When considering equality, this will be done by comparing all the fields
    of the node. This include even the `parent` and `branches` fields. This
    means two nodes are considered equal if they look the same and they are
    connected in to nodes that also look the same.
"""  # noqa: E501
from __future__ import annotations

import inspect
from collections.abc import Callable, Iterator, Mapping, Sequence
from dataclasses import dataclass, field
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Concatenate,
    Generic,
    Literal,
    NamedTuple,
    ParamSpec,
    TypeAlias,
    TypeVar,
    overload,
)
from typing_extensions import override

from more_itertools import first_true
from rich.text import Text
from sklearn.pipeline import Pipeline as SklearnPipeline

from amltk._functional import classname, mapping_select, prefix_keys
from amltk._richutil import RichRenderable
from amltk.exceptions import RequestNotMetError
from amltk.types import Config, Item, Space

if TYPE_CHECKING:
    from typing_extensions import Self

    from ConfigSpace import ConfigurationSpace
    from rich.console import RenderableType
    from rich.panel import Panel

    from amltk.pipeline.components import Choice, Join, Sequential
    from amltk.pipeline.parsers.optuna import OptunaSearchSpace

    NodeLike: TypeAlias = (
        set["Node" | "NodeLike"]
        | tuple["Node" | "NodeLike", ...]
        | list["Node" | "NodeLike"]
        | Callable[..., Item]
        | Item
    )

    SklearnPipelineT = TypeVar("SklearnPipelineT", bound=SklearnPipeline)

T = TypeVar("T")
ParserOutput = TypeVar("ParserOutput")
BuilderOutput = TypeVar("BuilderOutput")
P = ParamSpec("P")


_NotSet = object()


class RichOptions(NamedTuple):
    """Options for rich printing."""

    panel_color: str = "default"
    node_orientation: Literal["horizontal", "vertical"] = "horizontal"


@dataclass(frozen=True)
class ParamRequest(Generic[T]):
    """A parameter request for a node. This is most useful for things like seeds."""

    key: str
    """The key to request under."""

    default: T | object = _NotSet
    """The default value to use if the key is not found.

    If left as `_NotSet` (default) then an error will be raised if the
    parameter is not found during configuration with
    [`configure()`][amltk.pipeline.node.Node.configure].
    """

    @property
    def has_default(self) -> bool:
        """Whether this request has a default value."""
        return self.default is not _NotSet


def request(key: str, default: T | object = _NotSet) -> ParamRequest[T]:
    """Create a new parameter request.

    Args:
        key: The key to request under.
        default: The default value to use if the key is not found.
            If left as `_NotSet` (default) then the key will be removed from the
            config once [`configure`][amltk.pipeline.Node.configure] is called and
            nothing has been provided.
    """
    return ParamRequest(key=key, default=default)


@dataclass(init=False, frozen=True, eq=True)
class Node(RichRenderable, Generic[Item, Space]):
    """The core node class for the pipeline.

    These are simple objects that are named and linked together to form
    a chain. They are then wrapped in a `Pipeline` object to provide
    a convenient interface for interacting with the chain.
    """

    name: str = field(hash=True)
    """Name of the node"""

    item: Callable[..., Item] | Item | None = field(hash=False)
    """The item attached to this node"""

    nodes: tuple[Node, ...] = field(hash=True)
    """The nodes that this node leads to."""

    config: Config | None = field(hash=False)
    """The configuration for this node"""

    space: Space | None = field(hash=False)
    """The search space for this node"""

    fidelities: Mapping[str, Any] | None = field(hash=False)
    """The fidelities for this node"""

    config_transform: Callable[[Config, Any], Config] | None = field(hash=False)
    """A function that transforms the configuration of this node"""

    meta: Mapping[str, Any] | None = field(hash=False)
    """Any meta information about this node"""

    _NODES_INIT: ClassVar[Literal["args", "kwargs"] | None] = "args"
    """Whether __init__ takes nodes as positional args, kwargs or does not accept it."""

    RICH_OPTIONS: ClassVar[RichOptions] = RichOptions(
        panel_color="default",
        node_orientation="horizontal",
    )
    """Options for rich printing"""

    def __init__(
        self,
        *nodes: Node,
        name: str,
        item: Item | Callable[[Item], Item] | None = None,
        config: Config | None = None,
        space: Space | None = None,
        fidelities: Mapping[str, Any] | None = None,
        config_transform: Callable[[Config, Any], Config] | None = None,
        meta: Mapping[str, Any] | None = None,
    ):
        """Initialize a choice."""
        super().__init__()
        object.__setattr__(self, "name", name)
        object.__setattr__(self, "item", item)
        object.__setattr__(self, "config", config)
        object.__setattr__(self, "space", space)
        object.__setattr__(self, "fidelities", fidelities)
        object.__setattr__(self, "config_transform", config_transform)
        object.__setattr__(self, "meta", meta)
        object.__setattr__(self, "nodes", nodes)

    def __getitem__(self, key: str) -> Node:
        """Get the node with the given name."""
        found = first_true(
            self.nodes,
            None,
            lambda node: node.name == key,
        )
        if found is None:
            raise KeyError(
                f"Could not find node with name {key} in '{self.name}'."
                f" Available nodes are: {', '.join(node.name for node in self.nodes)}",
            )

        return found

    @override
    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented

        return (
            self.name == other.name
            and self.item == other.item
            and self.config == other.config
            and self.space == other.space
            and self.fidelities == other.fidelities
            and self.config_transform == other.config_transform
            and self.meta == other.meta
            and self.nodes == other.nodes
        )

    def __or__(self, other: Node | NodeLike) -> Choice:
        from amltk.pipeline.components import as_node

        return as_node({self, as_node(other)})

    def __and__(self, other: Node | NodeLike) -> Join:
        from amltk.pipeline.components import as_node

        return as_node((self, other))  # type: ignore

    def __rshift__(self, other: Node | NodeLike) -> Sequential:
        from amltk.pipeline.components import as_node

        return as_node([self, other])  # type: ignore

    def configure(
        self,
        config: Config,
        *,
        prefixed_name: bool | None = None,
        transform_context: Any | None = None,
        params: Mapping[str, Any] | None = None,
    ) -> Self:
        """Configure this node and anything following it with the given config.

        Args:
            config: The configuration to apply
            prefixed_name: Whether items in the config are prefixed by the names
                of the nodes.
                * If `None`, the default, then `prefixed_name` will be assumed to
                    be `True` if this node has a next node or if the config has
                    keys that begin with this nodes name.
                * If `True`, then the config will be searched for items prefixed
                    by the name of the node (and subsequent chained nodes).
                * If `False`, then the config will be searched for items without
                    the prefix, i.e. the config keys are exactly those matching
                    this nodes search space.
            transform_context: Any context to give to `config_transform=` of individual
                nodes.
            params: The params to match any requests when configuring this node.
                These will match against any ParamRequests in the config and will
                be used to fill in any missing values.

        Returns:
            The configured node
        """
        # Get the config for this node
        match prefixed_name:
            case True:
                config = mapping_select(config, f"{self.name}:")
            case False:
                pass
            case None if any(k.startswith(f"{self.name}:") for k in config):
                config = mapping_select(config, f"{self.name}:")
            case None:
                pass

        _kwargs: dict[str, Any] = {}

        # Configure all the branches if exists
        if len(self.nodes) > 0:
            nodes = tuple(
                node.configure(
                    config,
                    prefixed_name=True,
                    transform_context=transform_context,
                    params=params,
                )
                for node in self.nodes
            )
            _kwargs["nodes"] = nodes

        this_config = {
            hp: v
            for hp, v in config.items()
            if (
                ":" not in hp
                and not any(hp.startswith(f"{node.name}") for node in self.nodes)
            )
        }
        if self.config is not None:
            this_config = {**self.config, **this_config}

        this_config = dict(self._fufill_param_requests(this_config, params=params))

        if self.config_transform is not None:
            this_config = dict(self.config_transform(this_config, transform_context))

        if len(this_config) > 0:
            _kwargs["config"] = dict(this_config)

        return self.mutate(**_kwargs)

    def fidelity_space(self) -> dict[str, Any]:
        """Get the fidelities for this node and any connected nodes."""
        fids = {}
        for node in self.nodes:
            fids.update(prefix_keys(node.fidelity_space(), f"{self.name}:"))

        return fids

    def linearized_fidelity(self, value: float) -> dict[str, int | float | Any]:
        """Get the liniearized fidelities for this node and any connected nodes.

        Args:
            value: The value to linearize. Must be between [0, 1]

        Return:
            dictionary from key to it's linearized fidelity.
        """
        assert 1.0 <= value <= 100.0, f"{value=} not in [1.0, 100.0]"  # noqa: PLR2004
        d = {}
        for node in self.nodes:
            node_fids = prefix_keys(
                node.linearized_fidelity(value),
                f"{self.name}:",
            )
            d.update(node_fids)

        if self.fidelities is None:
            return d

        for f_name, f_range in self.fidelities.items():
            match f_range:
                case (int() | float(), int() | float()):
                    low, high = f_range
                    fid = low + (high - low) * value
                    fid = low + (high - low) * (value - 1) / 100
                    fid = fid if isinstance(low, float) else round(fid)
                    d[f_name] = fid
                case _:
                    raise ValueError(
                        f"Invalid fidelities to linearize {f_range} for {f_name}"
                        f" in {self}. Only supports ranges of the form (low, high)",
                    )

        return prefix_keys(d, f"{self.name}:")

    def iter(self) -> Iterator[Node]:
        """Iterate the the nodes, including this node.

        Yields:
            The nodes connected to this node
        """
        yield self
        for node in self.nodes:
            yield from node.iter()

    def mutate(self, **kwargs: Any) -> Self:
        """Mutate the node with the given keyword arguments.

        Args:
            **kwargs: The keyword arguments to mutate

        Returns:
            Self
                The mutated node
        """
        _args = ()
        _kwargs = {**self.__dict__, **kwargs}

        # If there's nodes in kwargs, we have to check if it's
        # a positional or keyword argument and handle accordingly.
        if (nodes := _kwargs.pop("nodes", None)) is not None:
            match self._NODES_INIT:
                case "args":
                    _args = nodes
                case "kwargs":
                    _kwargs["nodes"] = nodes
                case None if len(nodes) == 0:
                    pass  # Just ignore it, it's popped out
                case None:
                    raise ValueError(
                        "Cannot mutate nodes when __init__ does not accept nodes",
                    )

        # If there's a config in kwargs, we have to check if it's actually got values
        config = _kwargs.pop("config", None)
        if config is not None and len(config) > 0:
            _kwargs["config"] = config

        # Lastly, we remove anything that can't be passed to kwargs of the
        # subclasses __init__
        _available_kwargs = inspect.signature(self.__init__).parameters.keys()  # type: ignore
        for k in list(_kwargs.keys()):
            if k not in _available_kwargs:
                _kwargs.pop(k)

        return self.__class__(*_args, **_kwargs)

    def copy(self) -> Self:
        """Copy this node, removing all links in the process."""
        return self.mutate()

    def path_to(self, key: str | Node | Callable[[Node], bool]) -> list[Node] | None:
        """Find a path to the given node.

        Args:
            key: The key to search for or a function that returns True if the node
                is the desired node

        Returns:
            The path to the node if found, else None
        """
        # We found our target, just return now

        match key:
            case Node():
                pred = lambda node: node == key
            case str():
                pred = lambda node: node.name == key
            case _:
                pred = key

        for path, node in self.walk():
            if pred(node):
                return path

        return None

    def walk(
        self,
        path: Sequence[Node] | None = None,
    ) -> Iterator[tuple[list[Node], Node]]:
        """Walk the nodes in this chain.

        Args:
            path: The current path to this node

        Yields:
            The parents of the node and the node itself
        """
        path = list(path) if path is not None else []
        yield path, self

        for node in self.nodes:
            yield from node.walk(path=[*path, self])

    @overload
    def find(self, key: str | Node | Callable[[Node], bool], default: T) -> Node | T:
        ...

    @overload
    def find(self, key: str | Node | Callable[[Node], bool]) -> Node | None:
        ...

    def find(
        self,
        key: str | Node | Callable[[Node], bool],
        default: T | None = None,
    ) -> Node | T | None:
        """Find a node in that's nested deeper from this node.

        Args:
            key: The key to search for or a function that returns True if the node
                is the desired node
            default: The value to return if the node is not found. Defaults to None

        Returns:
            The node if found, otherwise the default value. Defaults to None
        """
        itr = self.iter()
        match key:
            case Node():
                return first_true(itr, default, lambda node: node == key)
            case str():
                return first_true(itr, default, lambda node: node.name == key)
            case _:
                return first_true(itr, default, key)  # type: ignore

    @overload
    def search_space(
        self,
        parser: Literal["configspace"],
        *,
        flat: bool = False,
        conditionals: bool = True,
        delim: str = ":",
    ) -> ConfigurationSpace:
        ...

    @overload
    def search_space(
        self,
        parser: Literal["optuna"],
        *,
        seed: int | None = None,
        flat: bool = False,
        conditionals: bool = True,
        delim: str = ":",
    ) -> OptunaSearchSpace:
        ...

    @overload
    def search_space(
        self,
        parser: Callable[Concatenate[Node, P], ParserOutput],
        *parser_args: P.args,
        **parser_kwargs: P.kwargs,
    ) -> ParserOutput:
        ...

    def search_space(
        self,
        parser: (
            Callable[Concatenate[Node, P], ParserOutput]
            | Literal["configspace", "optuna"]
        ),
        *parser_args: P.args,
        **parser_kwargs: P.kwargs,
    ) -> ParserOutput | ConfigurationSpace | OptunaSearchSpace:
        """Get the search space for this node."""
        match parser:
            case "configspace":
                from amltk.pipeline.parsers.configspace import parser as cs_parser

                return cs_parser(self, *parser_args, **parser_kwargs)  # type: ignore
            case "optuna":
                from amltk.pipeline.parsers.optuna import parser as optuna_parser

                return optuna_parser(self, *parser_args, **parser_kwargs)  # type: ignore
            case str():  # type: ignore
                raise ValueError(
                    f"Invalid str for parser {parser}. "
                    "Please use 'configspace' or 'optuna' or pass in your own"
                    " parser function",
                )
            case _:
                return parser(self, *parser_args, **parser_kwargs)

    @overload
    def build(
        self,
        builder: Literal["sklearn"],
        *builder_args: Any,
        pipeline_type: type[SklearnPipelineT] = SklearnPipeline,
        **builder_kwargs: Any,
    ) -> SklearnPipelineT:
        ...

    @overload
    def build(
        self,
        builder: Literal["sklearn"],
        *builder_args: Any,
        **builder_kwargs: Any,
    ) -> SklearnPipeline:
        ...

    @overload
    def build(
        self,
        builder: Callable[Concatenate[Node, P], BuilderOutput],
        *builder_args: P.args,
        **builder_kwargs: P.kwargs,
    ) -> BuilderOutput:
        ...

    def build(
        self,
        builder: Callable[Concatenate[Node, P], BuilderOutput] | Literal["sklearn"],
        *builder_args: P.args,
        **builder_kwargs: P.kwargs,
    ) -> BuilderOutput | SklearnPipeline:
        """Build a concrete object out of this node.

        Args:
            builder: The builder to use. This can be a function that takes in
                the node and returns the object or a string that is one of:

                * `#!python "sklearn"`: Build a
                    [`sklearn.pipeline.Pipeline`][sklearn.pipeline.Pipeline]
                    out of this node.

            builder_args: The positional arguments to pass to the builder
            builder_kwargs: The keyword arguments to pass to the builder

        Returns:
            The built object
        """
        match builder:
            case "sklearn":
                from amltk.pipeline.builders.sklearn import build as _build

                return _build(self, *builder_args, **builder_kwargs)  # type: ignore
            case _:
                return builder(self, *builder_args, **builder_kwargs)

    def _rich_iter(self) -> Iterator[RenderableType]:
        """Iterate the panels for rich printing."""
        yield self.__rich__()
        for node in self.nodes:
            yield from node._rich_iter()

    def _rich_table_items(self) -> Iterator[tuple[RenderableType, ...]]:
        """Get the items to add to the rich table."""
        from rich.pretty import Pretty

        from amltk._richutil import Function

        if self.item is not None:
            if isinstance(self.item, type) or callable(self.item):
                yield "item", Function(self.item, signature="...")
            else:
                yield "item", Pretty(self.item)

        if self.config is not None:
            yield "config", Pretty(self.config)

        if self.space is not None:
            yield "space", Pretty(self.space)

        if self.fidelities is not None:
            yield "fidelity", Pretty(self.fidelities)

        if self.config_transform is not None:
            yield "transform", Function(self.config_transform, signature="...")

        if self.meta is not None:
            yield "meta", Pretty(self.meta)

    def _rich_panel_contents(self) -> Iterator[RenderableType]:
        from rich.table import Table
        from rich.text import Text

        options = self.RICH_OPTIONS

        if panel_contents := list(self._rich_table_items()):
            table = Table.grid(padding=(0, 1), expand=False)
            for tup in panel_contents:
                table.add_row(*tup, style="default")
            table.add_section()
            yield table

        if len(self.nodes) > 0:
            match options:
                case RichOptions(node_orientation="horizontal"):
                    table = Table.grid(padding=(0, 1), expand=False)
                    nodes = [node.__rich__() for node in self.nodes]
                    table.add_row(*nodes)
                    yield table
                case RichOptions(node_orientation="vertical"):
                    first, *rest = self.nodes
                    yield first.__rich__()
                    for node in rest:
                        yield Text("↓", style="bold", justify="center")
                        yield node.__rich__()
                case _:
                    raise ValueError(f"Invalid orientation {options.node_orientation}")

    def display(self, *, full: bool = False) -> RenderableType:
        """Display this node.

        Args:
            full: Whether to display the full node or just a summary
        """
        if not full:
            return self.__rich__()

        from rich.console import Group as RichGroup

        return RichGroup(*self._rich_iter())

    @override
    def __rich__(self) -> Panel:
        from rich.console import Group as RichGroup
        from rich.panel import Panel

        clr = self.RICH_OPTIONS.panel_color
        title = Text.assemble(
            (classname(self), f"{clr} bold"),
            "(",
            (self.name, f"{clr} italic"),
            ")",
            style="default",
            end="",
        )
        contents = list(self._rich_panel_contents())
        _content = contents[0] if len(contents) == 1 else RichGroup(*contents)
        return Panel(
            _content,
            title=title,
            title_align="left",
            border_style=clr,
            expand=False,
        )

    def _fufill_param_requests(
        self,
        config: Config,
        params: Mapping[str, Any] | None = None,
    ) -> Config:
        _params = params or {}
        new_config = dict(config)

        for k, request in config.items():
            match request:
                case ParamRequest(key=request_key) if request_key in _params:
                    new_config[k] = _params[request_key]
                case ParamRequest(default=default) if request.has_default:
                    new_config[k] = default
                case ParamRequest():
                    raise RequestNotMetError(
                        f"Missing {request=} for {self}.\nparams={params}",
                    )
                case _:
                    continue

        return new_config



"""Conversions between different data repesentations and formats."""
from __future__ import annotations

from typing import TYPE_CHECKING, overload

import numpy as np
import pandas as pd

if TYPE_CHECKING:
    import numpy.typing as npt


def probabilities_to_classes(
    probabilities: npt.NDArray[np.floating],
    classes: np.ndarray | npt.ArrayLike | list,
) -> np.ndarray:
    """Convert probabilities to classes.

    !!! note

        Converts using the logic of `predict()` of `RandomForestClassifier`.

    Args:
        probabilities: The probabilities to convert
        classes: The classes to use.

    Returns:
        The classes corresponding to the probabilities
    """
    # Taken from `predict()` of RandomForestclassifier
    classes = np.asarray(classes)
    n_outputs = 1 if classes.ndim == 1 else classes.shape[1]
    if n_outputs == 1:
        return classes.take(np.argmax(probabilities, axis=1), axis=0)  # type: ignore

    n_samples = probabilities[0].shape[0]
    # all dtypes should be the same, so just take the first
    class_type = classes[0].dtype
    predictions = np.empty((n_samples, n_outputs), dtype=class_type)

    for k in range(n_outputs):
        predictions[:, k] = classes[k].take(np.argmax(probabilities[k], axis=1), axis=0)

    return predictions


def to_numpy(
    x: np.ndarray | pd.DataFrame | pd.Series,
    *,
    flatten_if_1d: bool = False,
) -> np.ndarray:
    """Convert to numpy array.

    Args:
        x: The data to convert
        flatten_if_1d: Whether to flatten the array if it is 1d

    Returns:
        The converted data
    """
    _x = x.to_numpy() if isinstance(x, pd.DataFrame | pd.Series) else np.asarray(x)

    if (
        flatten_if_1d
        and x.ndim == 2  # noqa: PLR2004 # type: ignore
        and x.shape[1] == 1  # type: ignore
    ):
        _x = np.ravel(_x)

    assert isinstance(_x, np.ndarray)
    return _x


@overload
def flatten_if_1d(x: pd.DataFrame | pd.Series) -> pd.DataFrame | pd.Series:
    ...


@overload
def flatten_if_1d(x: np.ndarray) -> np.ndarray:  # type: ignore
    ...


def flatten_if_1d(
    x: np.ndarray | pd.DataFrame | pd.Series,
) -> np.ndarray | pd.DataFrame | pd.Series:
    """Flatten if 1d.

    Retains the type of the input, i.e. pandas stays pandas and numpy stays numpy.

    Args:
        x: The data to flatten

    Returns:
        The flattened data
    """
    if isinstance(x, np.ndarray) and x.ndim == 2 and x.shape[1] == 1:  # noqa: PLR2004
        x = np.ravel(x)
    elif (
        isinstance(x, pd.DataFrame) and x.ndim == 2 and x.shape[1] == 1  # noqa: PLR2004
    ):
        x = x.iloc[:, 0]

    return x


def is_str_object_dtype(x: np.ndarray) -> bool:
    """Check if object dtype and string values.

    Args:
        x: The data to check

    Returns:
        Whether it is object dtype and string values
    """
    return x.dtype == object and isinstance(x[0], str)


def as_str_dtype_if_str_object(x: np.ndarray) -> np.ndarray:
    """Convert to string dtype if object dtype and string values.

    Args:
        x: The data to convert

    Returns:
        The converted data if it can be done
    """
    if is_str_object_dtype(x):
        return x.astype(str)
    return x



"""A [`Task`][amltk.scheduling.task.Task] is a unit of work that can be scheduled by the
[`Scheduler`][amltk.scheduling.Scheduler].

It is defined by its `function=` to call. Whenever a `Task`
has its [`submit()`][amltk.scheduling.task.Task.submit] method called,
the function will be dispatched to run by a `Scheduler`.

When a task has returned, either successfully, or with an exception,
it will emit `@events` to indicate so. You can subscribe to these events
with callbacks and act accordingly.


??? example "`@events`"

    Check out the `@events` reference
    for more on how to customize these callbacks. You can also take a look
    at the API of [`on()`][amltk.scheduling.task.Task.on] for more information.

    === "`@on_result`"

        ::: amltk.scheduling.task.Task.on_result

    === "`@on_exception`"

        ::: amltk.scheduling.task.Task.on_exception

    === "`@on_done`"

        ::: amltk.scheduling.task.Task.on_done

    === "`@on_submitted`"

        ::: amltk.scheduling.task.Task.on_submitted

    === "`@on_cancelled`"

        ::: amltk.scheduling.task.Task.on_cancelled

??? tip "Usage"

    The usual way to create a task is with
    [`Scheduler.task()`][amltk.scheduling.scheduler.Scheduler.task],
    where you provide the `function=` to call.

    ```python exec="true" source="material-block" html="true"
    from amltk import Scheduler

    def f(x: int) -> int:
        return x * 2
    from amltk._doc import make_picklable; make_picklable(f)  # markdown-exec: hide

    scheduler = Scheduler.with_processes(2)
    task = scheduler.task(f)

    @scheduler.on_start
    def on_start():
        task.submit(1)

    @task.on_result
    def on_result(future: Future[int], result: int):
        print(f"Task {future} returned {result}")

    scheduler.run()
    from amltk._doc import doc_print; doc_print(print, scheduler)  # markdown-exec: hide
    ```

    If you'd like to simply just call the original function, without submitting it to
    the scheduler, you can always just call the task directly, i.e. `#!python task(1)`.

You can also provide [`Plugins`][amltk.scheduling.plugins.Plugin] to the task,
to modify tasks, add functionality and add new events.
"""
from __future__ import annotations

import logging
from asyncio import Future
from collections.abc import Callable, Iterable
from typing import TYPE_CHECKING, Any, Concatenate, Generic, TypeVar, overload
from typing_extensions import ParamSpec, Self, override

from more_itertools import first_true

from amltk._functional import callstring
from amltk._richutil.renderable import RichRenderable
from amltk.exceptions import EventNotKnownError, SchedulerNotRunningError
from amltk.randomness import randuid
from amltk.scheduling.events import Emitter, Event, Subscriber
from amltk.scheduling.plugins.plugin import Plugin

if TYPE_CHECKING:
    from rich.panel import Panel

    from amltk.scheduling.scheduler import Scheduler

logger = logging.getLogger(__name__)


P = ParamSpec("P")
P2 = ParamSpec("P2")

R = TypeVar("R")
R2 = TypeVar("R2")
CallableT = TypeVar("CallableT", bound=Callable)


class Task(RichRenderable, Generic[P, R]):
    """The task class."""

    unique_ref: str
    """A unique reference to this task."""
    plugins: list[Plugin]
    """The plugins to use for this task."""
    function: Callable[P, R]
    """The function of this task"""
    scheduler: Scheduler
    """The scheduler that this task is registered with."""
    init_plugins: bool
    """Whether to initialize the plugins or not."""
    queue: list[Future[R]]
    """The queue of futures for this task."""
    emitter: Emitter
    """The emitter for events of this task."""

    on_submitted: Subscriber[Concatenate[Future[R], P]]
    """An event that is emitted when a future is submitted to the
    scheduler. It will pass the future as the first argument with args and
    kwargs following.

    This is done before any callbacks are attached to the future.
    ```python
    @task.on_submitted
    def on_submitted(future: Future[R], *args, **kwargs):
        print(f"Future {future} was submitted with {args=} and {kwargs=}")
    ```
    """
    on_done: Subscriber[Future[R]]
    """Called when a task is done running with a result or exception.
    ```python
    @task.on_done
    def on_done(future: Future[R]):
        print(f"Future {future} is done")
    ```
    """
    on_cancelled: Subscriber[Future[R]]
    """Called when a task is cancelled.
    ```python
    @task.on_cancelled
    def on_cancelled(future: Future[R]):
        print(f"Future {future} was cancelled")
    ```
    """
    on_result: Subscriber[Future[R], R]
    """Called when a task has successfully returned a value.
    Comes with Future
    ```python
    @task.on_result
    def on_result(future: Future[R], result: R):
        print(f"Future {future} returned {result}")
    ```
    """
    on_exception: Subscriber[Future[R], BaseException]
    """Called when a task failed to return anything but an exception.
    Comes with Future
    ```python
    @task.on_exception
    def on_exception(future: Future[R], error: BaseException):
        print(f"Future {future} exceptioned {error}")
    ```
    """

    SUBMITTED: Event[Concatenate[Future[R], P]] = Event("on_submitted")
    DONE: Event[Future[R]] = Event("on_done")

    CANCELLED: Event[Future[R]] = Event("on_cancelled")
    RESULT: Event[Future[R], R] = Event("on_result")
    EXCEPTION: Event[Future[R], BaseException] = Event("on_exception")

    def __init__(
        self: Self,
        function: Callable[P, R],
        scheduler: Scheduler,
        *,
        plugins: Plugin | Iterable[Plugin] = (),
        init_plugins: bool = True,
    ) -> None:
        """Initialize a task.

        Args:
            function: The function of this task
            scheduler: The scheduler that this task is registered with.
            plugins: The plugins to use for this task.
            init_plugins: Whether to initialize the plugins or not.
        """
        super().__init__()
        self.unique_ref = randuid(8)

        self.emitter = Emitter()
        self.event_counts = self.emitter.event_counts
        self.plugins: list[Plugin] = (
            [plugins] if isinstance(plugins, Plugin) else list(plugins)
        )
        self.function: Callable[P, R] = function
        self.scheduler: Scheduler = scheduler
        self.init_plugins: bool = init_plugins
        self.queue: list[Future[R]] = []

        # Set up subscription methods to events
        self.on_submitted = self.emitter.subscriber(self.SUBMITTED)
        self.on_done = self.emitter.subscriber(self.DONE)
        self.on_result = self.emitter.subscriber(self.RESULT)
        self.on_exception = self.emitter.subscriber(self.EXCEPTION)
        self.on_cancelled = self.emitter.subscriber(self.CANCELLED)

        if init_plugins:
            for plugin in self.plugins:
                plugin.attach_task(self)

    def futures(self) -> list[Future[R]]:
        """Get the futures for this task.

        Returns:
            A list of futures for this task.
        """
        return self.queue

    @overload
    def on(
        self,
        event: Event[P2],
        callback: None = None,
        *,
        when: Callable[[], bool] | None = ...,
        limit: int | None = ...,
        repeat: int = ...,
        every: int = ...,
    ) -> Subscriber[P2]:
        ...

    @overload
    def on(
        self,
        event: str,
        callback: None = None,
        *,
        when: Callable[[], bool] | None = ...,
        limit: int | None = ...,
        repeat: int = ...,
        every: int = ...,
    ) -> Subscriber[...]:
        ...

    @overload
    def on(
        self,
        event: str,
        callback: Callable,
        *,
        when: Callable[[], bool] | None = ...,
        limit: int | None = ...,
        repeat: int = ...,
        every: int = ...,
    ) -> None:
        ...

    def on(
        self,
        event: Event[P2] | str,
        callback: Callable[P2, Any] | None = None,
        *,
        when: Callable[[], bool] | None = None,
        limit: int | None = None,
        repeat: int = 1,
        every: int = 1,
        hidden: bool = False,
    ) -> Subscriber[P2] | Subscriber[...] | None:
        """Subscribe to an event.

        Args:
            event: The event to subscribe to.
            callback: The callback to call when the event is emitted.
                If not specified, what is returned can be used as a decorator.
            when: A predicate to determine whether to call the callback.
            limit: The number of times to call the callback.
            repeat: The number of times to repeat the subscription.
            every: The number of times to wait between repeats.
            hidden: Whether to hide the callback in visual output.
                This is mainly used to facilitate Plugins who
                act upon events but don't want to be seen, primarily
                as they are just book-keeping callbacks.

        Returns:
            The subscriber if no callback was provided, otherwise `None`.
        """
        if isinstance(event, str):
            _e = first_true(self.emitter.events, None, lambda e: e.name == event)
            if _e is None:
                raise EventNotKnownError(
                    f"{event=} is not a valid event."
                    f"\nKnown events are: {[e.name for e in self.emitter.events]}",
                )
        else:
            _e = event

        subscriber = self.emitter.subscriber(
            _e,  # type: ignore
            when=when,
            limit=limit,
            repeat=repeat,
            every=every,
        )
        if callback is None:
            return subscriber

        subscriber(callback, hidden=hidden)
        return None

    @property
    def n_running(self) -> int:
        """Get the number of futures for this task that are currently running."""
        return sum(1 for f in self.queue if not f.done())

    def running(self) -> bool:
        """Check if this task has any futures that are currently running."""
        return self.n_running > 0

    def submit(self, *args: P.args, **kwargs: P.kwargs) -> Future[R] | None:
        """Dispatch this task.

        Args:
            *args: The positional arguments to pass to the task.
            **kwargs: The keyword arguments to call the task with.

        Returns:
            The future of the task, or `None` if the limit was reached.

        Raises:
            SchedulerNotRunningError: If the scheduler is not running.
                You can protect against this using,
                [`scheduler.running()`][amltk.scheduling.scheduler.Scheduler.running].
        """
        # Inform all plugins that the task is about to be called
        # They have chance to cancel submission based on their return
        # value.
        fn = self.function
        for plugin in self.plugins:
            items = plugin.pre_submit(fn, *args, **kwargs)
            if items is None:
                logger.debug(
                    f"Plugin '{plugin.name}' prevented {self} from being submitted"
                    f" with {callstring(self.function, *args, **kwargs)}",
                )
                return None

            fn, args, kwargs = items  # type: ignore

        try:
            future = self.scheduler.submit(fn, *args, **kwargs)
        except SchedulerNotRunningError as e:
            logger.exception("Scheduler is not running", exc_info=e)
            raise e
        except Exception as e:
            logger.exception("Error submitting task", exc_info=e)
            raise e

        self.queue.append(future)

        # We have the function wrapped in something will
        # attach tracebacks to errors, so we need to get the
        # original function name.
        msg = f"Submitted {callstring(self.function, *args, **kwargs)} from {self}."
        logger.debug(msg)
        self.on_submitted.emit(future, *args, **kwargs)

        # Process the task once it's completed
        # NOTE: If the task is done super quickly or in the sequential mode,
        # this will immediatly call `self._process_future`.
        future.add_done_callback(self._process_future)
        return future

    def copy(self, *, init_plugins: bool = True) -> Self:
        """Create a copy of this task.

        Will use the same scheduler and function, but will have a different
        event manager such that any events listend to on the old task will
        **not** trigger with the copied task.

        Args:
            init_plugins: Whether to initialize the copied plugins on the copied
                task. Usually you will want to leave this as `True`.

        Returns:
            A copy of this task.
        """
        return self.__class__(
            self.function,
            self.scheduler,
            plugins=tuple(p.copy() for p in self.plugins),
            init_plugins=init_plugins,
        )

    def _process_future(self, future: Future[R]) -> None:
        try:
            self.queue.remove(future)
        except ValueError as e:
            raise ValueError(f"{future=} not found in task queue {self.queue=}") from e

        if future.cancelled():
            self.on_cancelled.emit(future)
            return

        self.on_done.emit(future)

        exception = future.exception()
        if exception is not None:
            self.on_exception.emit(future, exception)
        else:
            result = future.result()
            self.on_result.emit(future, result)

    def attach_plugin(self, plugin: Plugin) -> None:
        """Attach a plugin to this task.

        Args:
            plugin: The plugin to attach.
        """
        self.plugins.append(plugin)
        plugin.attach_task(self)

    def _when_future_from_submission(
        self,
        future: Future[R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        """Access the future before the callbacks for the future are registered.

        This is primarly to allow subclasses of Task to know of the future obtained
        from submitting to the Scheduler, **before** the callbacks are registered.
        This can be required in the case the task finishes super quickly, meaning
        callbacks are registered before the subtask can do anything. This also
        necessarily happens in a sequential execution Scheduler.
        """

    @override
    def __repr__(self) -> str:
        kwargs = {"unique_ref": self.unique_ref}
        kwargs_str = ", ".join(f"{k}={v}" for k, v in kwargs.items())
        return f"{self.__class__.__name__}({kwargs_str})"

    @override
    def __rich__(self) -> Panel:
        from rich.console import Group as RichGroup
        from rich.panel import Panel
        from rich.text import Text
        from rich.tree import Tree

        from amltk._richutil import Function

        items: list[RichRenderable | Tree] = []

        if any(self.plugins):
            for plugin in self.plugins:
                items.append(plugin)

        tree = Tree(label="", hide_root=True)
        tree.add(self.emitter)
        items.append(tree)

        return Panel(
            RichGroup(*items),
            title=Function(self.function, prefix="Task").__rich__(),
            title_align="left",
            border_style="deep_sky_blue2",
            subtitle=Text("Ref: ").append(self.unique_ref, "yellow italic"),
        )



"""You can use the various different node types to build a pipeline.

You can connect these nodes together using either the constructors explicitly,
as shown in the examples. We also provide some index operators:

* `>>` - Connect nodes together to form a [`Sequential`][amltk.pipeline.components.Sequential]
* `&` - Connect nodes together to form a [`Join`][amltk.pipeline.components.Join]
* `|` - Connect nodes together to form a [`Choice`][amltk.pipeline.components.Choice]

There is also another short-hand that you may find useful to know:

* `{comp1, comp2, comp3}` - This will automatically be converted into a
    [`Choice`][amltk.pipeline.Choice] between the given components.
* `(comp1, comp2, comp3)` - This will automatically be converted into a
    [`Join`][amltk.pipeline.Join] between the given components.
* `[comp1, comp2, comp3]` - This will automatically be converted into a
    [`Sequential`][amltk.pipeline.Sequential] between the given components.

For each of these components we will show examples using
the [`#! "sklearn"` builder][amltk.pipeline.builders.sklearn.build]

The components are:

### Component

::: amltk.pipeline.components.Component
    options:
        members: false

### Sequential

::: amltk.pipeline.components.Sequential
    options:
        members: false

### Choice

::: amltk.pipeline.components.Choice
    options:
        members: false

### Split

::: amltk.pipeline.components.Split
    options:
        members: false

### Join

::: amltk.pipeline.components.Join
    options:
        members: false

### Fixed

::: amltk.pipeline.components.Fixed
    options:
        members: false

### Searchable

::: amltk.pipeline.components.Searchable
    options:
        members: false
"""  # noqa: E501
from __future__ import annotations

import inspect
from collections.abc import Callable, Iterator, Mapping, Sequence
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, ClassVar, TypeVar, overload
from typing_extensions import override

from more_itertools import all_unique, first_true

from amltk._functional import entity_name
from amltk.exceptions import DuplicateNamesError, NoChoiceMadeError, NodeNotFoundError
from amltk.pipeline.node import Node, RichOptions
from amltk.randomness import randuid
from amltk.types import Config, Item, Space

if TYPE_CHECKING:
    from amltk.pipeline.node import NodeLike


T = TypeVar("T")
NodeT = TypeVar("NodeT", bound=Node)


@overload
def as_node(thing: Node, name: str | None = ...) -> Node:  # type: ignore
    ...


@overload
def as_node(thing: tuple[Node | NodeLike, ...], name: str | None = ...) -> Join:  # type: ignore
    ...


@overload
def as_node(thing: set[Node | NodeLike], name: str | None = ...) -> Choice:  # type: ignore
    ...


@overload
def as_node(thing: list[Node | NodeLike], name: str | None = ...) -> Sequential:  # type: ignore
    ...


@overload
def as_node(  # type: ignore
    thing: Callable[..., Item],
    name: str | None = ...,
) -> Component[Item, None]:
    ...


@overload
def as_node(thing: Item, name: str | None = ...) -> Fixed[Item]:
    ...


def as_node(  # noqa: PLR0911
    thing: Node | NodeLike[Item],
    name: str | None = None,
) -> Node | Choice | Join | Sequential | Fixed[Item]:
    """Convert a node, pipeline, set or tuple into a component, copying anything
    in the process and removing all linking to other nodes.

    Args:
        thing: The thing to convert
        name: The name of the node. If it already a node, it will be renamed to that
            one.

    Returns:
        The component
    """
    match thing:
        case set():
            return Choice(*thing, name=name)
        case tuple():
            return Join(*thing, name=name)
        case list():
            return Sequential(*thing, name=name)
        case Node():
            name = thing.name if name is None else name
            return thing.mutate(name=name)
        case type():
            return Component(thing, name=name)
        case thing if (inspect.isfunction(thing) or inspect.ismethod(thing)):
            return Component(thing, name=name)
        case _:
            return Fixed(thing, name=name)


@dataclass(init=False, frozen=True, eq=True)
class Join(Node[Item, Space]):
    """[`Join`][amltk.pipeline.Join] together different parts of the pipeline.

    This indicates the different children in
    [`.nodes`][amltk.pipeline.Node.nodes] should act in tandem with one
    another, for example, concatenating the outputs of the various members of the
    `Join`.

    ```python exec="true" source="material-block" html="true"
    from amltk.pipeline import Join, Component
    from sklearn.decomposition import PCA
    from sklearn.feature_selection import SelectKBest

    pca = Component(PCA, space={"n_components": (1, 3)})
    kbest = Component(SelectKBest, space={"k": (1, 3)})

    join = Join(pca, kbest, name="my_feature_union")
    from amltk._doc import doc_print; doc_print(print, join)  # markdown-exec: hide

    space = join.search_space("configspace")
    from amltk._doc import doc_print; doc_print(print, space)  # markdown-exec: hide

    pipeline = join.build("sklearn")
    from amltk._doc import doc_print; doc_print(print, pipeline)  # markdown-exec: hide
    ```

    You may also just join together nodes using an infix operator `&` if you prefer:

    ```python exec="true" source="material-block" html="true"
    from amltk.pipeline import Join, Component
    from sklearn.decomposition import PCA
    from sklearn.feature_selection import SelectKBest

    pca = Component(PCA, space={"n_components": (1, 3)})
    kbest = Component(SelectKBest, space={"k": (1, 3)})

    # Can not parametrize or name the join
    join = pca & kbest
    from amltk._doc import doc_print; doc_print(print, join)  # markdown-exec: hide

    # With a parametrized join
    join = (
        Join(name="my_feature_union") & pca & kbest
    )
    item = join.build("sklearn")
    print(item._repr_html_())  # markdown-exec: hide
    ```

    Whenever some other node sees a tuple, i.e. `(comp1, comp2, comp3)`, this
    will automatically be converted into a `Join`.

    ```python exec="true" source="material-block" html="true"
    from amltk.pipeline import Sequential, Component
    from sklearn.decomposition import PCA
    from sklearn.feature_selection import SelectKBest
    from sklearn.ensemble import RandomForestClassifier

    pca = Component(PCA, space={"n_components": (1, 3)})
    kbest = Component(SelectKBest, space={"k": (1, 3)})

    # Can not parametrize or name the join
    join = Sequential(
        (pca, kbest),
        RandomForestClassifier(n_estimators=5),
        name="my_feature_union",
    )
    print(join._repr_html_())  # markdown-exec: hide
    ```

    Like all [`Node`][amltk.pipeline.node.Node]s, a `Join` accepts an explicit
    [`name=`][amltk.pipeline.node.Node.name],
    [`item=`][amltk.pipeline.node.Node.item],
    [`config=`][amltk.pipeline.node.Node.config],
    [`space=`][amltk.pipeline.node.Node.space],
    [`fidelities=`][amltk.pipeline.node.Node.fidelities],
    [`config_transform=`][amltk.pipeline.node.Node.config_transform] and
    [`meta=`][amltk.pipeline.node.Node.meta].

    See Also:
        * [`Node`][amltk.pipeline.node.Node]
    """

    nodes: tuple[Node, ...]
    """The nodes that this node leads to."""

    RICH_OPTIONS: ClassVar[RichOptions] = RichOptions(panel_color="#7E6B8F")

    _NODES_INIT: ClassVar = "args"

    def __init__(
        self,
        *nodes: Node | NodeLike,
        name: str | None = None,
        item: Item | Callable[[Item], Item] | None = None,
        config: Config | None = None,
        space: Space | None = None,
        fidelities: Mapping[str, Any] | None = None,
        config_transform: Callable[[Config, Any], Config] | None = None,
        meta: Mapping[str, Any] | None = None,
    ):
        """See [`Node`][amltk.pipeline.node.Node] for details."""
        _nodes = tuple(as_node(n) for n in nodes)
        if not all_unique(_nodes, key=lambda node: node.name):
            raise ValueError(
                f"Can't handle nodes they do not all contain unique names, {nodes=}."
                "\nAll nodes must have a unique name. Please provide a `name=` to them",
            )

        if name is None:
            name = f"Join-{randuid(8)}"

        super().__init__(
            *_nodes,
            name=name,
            item=item,
            config=config,
            space=space,
            fidelities=fidelities,
            config_transform=config_transform,
            meta=meta,
        )

    @override
    def __and__(self, other: Node | NodeLike) -> Join:
        other_node = as_node(other)
        if any(other_node.name == node.name for node in self.nodes):
            raise ValueError(
                f"Can't handle node with name '{other_node.name} as"
                f" there is already a node called '{other_node.name}' in {self.name}",
            )

        nodes = (*tuple(as_node(n) for n in self.nodes), other_node)
        return self.mutate(name=self.name, nodes=nodes)


@dataclass(init=False, frozen=True, eq=True)
class Choice(Node[Item, Space]):
    """A [`Choice`][amltk.pipeline.Choice] between different subcomponents.

    This indicates that a choice should be made between the different children in
    [`.nodes`][amltk.pipeline.Node.nodes], usually done when you
    [`configure()`][amltk.pipeline.node.Node.configure] with some `config` from
    a [`search_space()`][amltk.pipeline.node.Node.search_space].

    ```python exec="true" source="material-block" html="true"
    from amltk.pipeline import Choice, Component
    from sklearn.ensemble import RandomForestClassifier
    from sklearn.neural_network import MLPClassifier

    rf = Component(RandomForestClassifier, space={"n_estimators": (10, 100)})
    mlp = Component(MLPClassifier, space={"activation": ["logistic", "relu", "tanh"]})

    estimator_choice = Choice(rf, mlp, name="estimator")
    from amltk._doc import doc_print; doc_print(print, estimator_choice)  # markdown-exec: hide

    space = estimator_choice.search_space("configspace")
    from amltk._doc import doc_print; doc_print(print, space)  # markdown-exec: hide

    config = space.sample_configuration()
    from amltk._doc import doc_print; doc_print(print, config)  # markdown-exec: hide

    configured_choice = estimator_choice.configure(config)
    from amltk._doc import doc_print; doc_print(print, configured_choice)  # markdown-exec: hide

    chosen_estimator = configured_choice.chosen()
    from amltk._doc import doc_print; doc_print(print, chosen_estimator)  # markdown-exec: hide

    estimator = chosen_estimator.build_item()
    from amltk._doc import doc_print; doc_print(print, estimator)  # markdown-exec: hide
    ```

    You may also just add nodes to a `Choice` using an infix operator `|` if you prefer:

    ```python exec="true" source="material-block" html="true"
    from amltk.pipeline import Choice, Component
    from sklearn.ensemble import RandomForestClassifier
    from sklearn.neural_network import MLPClassifier

    rf = Component(RandomForestClassifier, space={"n_estimators": (10, 100)})
    mlp = Component(MLPClassifier, space={"activation": ["logistic", "relu", "tanh"]})

    estimator_choice = (
        Choice(name="estimator") | mlp | rf
    )
    from amltk._doc import doc_print; doc_print(print, estimator_choice)  # markdown-exec: hide
    ```

    Whenever some other node sees a set, i.e. `{comp1, comp2, comp3}`, this
    will automatically be converted into a `Choice`.

    ```python exec="true" source="material-block" html="true"
    from amltk.pipeline import Choice, Component, Sequential
    from sklearn.ensemble import RandomForestClassifier
    from sklearn.neural_network import MLPClassifier
    from sklearn.impute import SimpleImputer

    rf = Component(RandomForestClassifier, space={"n_estimators": (10, 100)})
    mlp = Component(MLPClassifier, space={"activation": ["logistic", "relu", "tanh"]})

    pipeline = Sequential(
        SimpleImputer(fill_value=0),
        {mlp, rf},
        name="my_pipeline",
    )
    from amltk._doc import doc_print; doc_print(print, pipeline)  # markdown-exec: hide
    ```

    Like all [`Node`][amltk.pipeline.node.Node]s, a `Choice` accepts an explicit
    [`name=`][amltk.pipeline.node.Node.name],
    [`item=`][amltk.pipeline.node.Node.item],
    [`config=`][amltk.pipeline.node.Node.config],
    [`space=`][amltk.pipeline.node.Node.space],
    [`fidelities=`][amltk.pipeline.node.Node.fidelities],
    [`config_transform=`][amltk.pipeline.node.Node.config_transform] and
    [`meta=`][amltk.pipeline.node.Node.meta].

    !!! warning "Order of nodes"

        The given nodes of a choice are always ordered according
        to their name, so indexing `choice.nodes` may not be reliable
        if modifying the choice dynamically.

        Please use `choice["name"]` to access the nodes instead.

    See Also:
        * [`Node`][amltk.pipeline.node.Node]
    """  # noqa: E501

    nodes: tuple[Node, ...]
    """The nodes that this node leads to."""

    RICH_OPTIONS: ClassVar[RichOptions] = RichOptions(panel_color="#FF4500")
    _NODES_INIT: ClassVar = "args"

    def __init__(
        self,
        *nodes: Node | NodeLike,
        name: str | None = None,
        item: Item | Callable[[Item], Item] | None = None,
        config: Config | None = None,
        space: Space | None = None,
        fidelities: Mapping[str, Any] | None = None,
        config_transform: Callable[[Config, Any], Config] | None = None,
        meta: Mapping[str, Any] | None = None,
    ):
        """See [`Node`][amltk.pipeline.node.Node] for details."""
        _nodes: tuple[Node, ...] = tuple(
            sorted((as_node(n) for n in nodes), key=lambda n: n.name),
        )
        if not all_unique(_nodes, key=lambda node: node.name):
            raise ValueError(
                f"Can't handle nodes as we can not generate a __choice__ for {nodes=}."
                "\nAll nodes must have a unique name. Please provide a `name=` to them",
            )

        if name is None:
            name = f"Choice-{randuid(8)}"

        super().__init__(
            *_nodes,
            name=name,
            item=item,
            config=config,
            space=space,
            fidelities=fidelities,
            config_transform=config_transform,
            meta=meta,
        )

    @override
    def __or__(self, other: Node | NodeLike) -> Choice:
        other_node = as_node(other)
        if any(other_node.name == node.name for node in self.nodes):
            raise ValueError(
                f"Can't handle node with name '{other_node.name} as"
                f" there is already a node called '{other_node.name}' in {self.name}",
            )

        nodes = tuple(
            sorted(
                [as_node(n) for n in self.nodes] + [other_node],
                key=lambda n: n.name,
            ),
        )
        return self.mutate(name=self.name, nodes=nodes)

    def chosen(self) -> Node:
        """The chosen branch.

        Returns:
            The chosen branch
        """
        match self.config:
            case {"__choice__": choice}:
                chosen = first_true(
                    self.nodes,
                    pred=lambda node: node.name == choice,
                    default=None,
                )
                if chosen is None:
                    raise NodeNotFoundError(choice, self.name)

                return chosen
            case _:
                raise NoChoiceMadeError(self.name)


@dataclass(init=False, frozen=True, eq=True)
class Sequential(Node[Item, Space]):
    """A [`Sequential`][amltk.pipeline.Sequential] set of operations in a pipeline.

    This indicates the different children in
    [`.nodes`][amltk.pipeline.Node.nodes] should act one after
    another, feeding the output of one into the next.

    ```python exec="true" source="material-block" html="true"
    from amltk.pipeline import Component, Sequential
    from sklearn.decomposition import PCA
    from sklearn.ensemble import RandomForestClassifier

    pipeline = Sequential(
        PCA(n_components=3),
        Component(RandomForestClassifier, space={"n_estimators": (10, 100)}),
        name="my_pipeline"
    )
    from amltk._doc import doc_print; doc_print(print, pipeline)  # markdown-exec: hide

    space = pipeline.search_space("configspace")
    from amltk._doc import doc_print; doc_print(print, space)  # markdown-exec: hide

    configuration = space.sample_configuration()
    from amltk._doc import doc_print; doc_print(print, configuration)  # markdown-exec: hide

    configured_pipeline = pipeline.configure(configuration)
    from amltk._doc import doc_print; doc_print(print, configured_pipeline)  # markdown-exec: hide

    sklearn_pipeline = pipeline.build("sklearn")
    print(sklearn_pipeline._repr_html_())  # markdown-exec: hide
    ```

    You may also just chain together nodes using an infix operator `>>` if you prefer:

    ```python exec="true" source="material-block" html="true"
    from amltk.pipeline import Join, Component, Sequential
    from sklearn.decomposition import PCA
    from sklearn.ensemble import RandomForestClassifier

    pipeline = (
        Sequential(name="my_pipeline")
        >> PCA(n_components=3)
        >> Component(RandomForestClassifier, space={"n_estimators": (10, 100)})
    )
    from amltk._doc import doc_print; doc_print(print, pipeline)  # markdown-exec: hide
    ```

    Whenever some other node sees a list, i.e. `[comp1, comp2, comp3]`, this
    will automatically be converted into a `Sequential`.

    ```python exec="true" source="material-block" html="true"
    from amltk.pipeline import Choice
    from sklearn.impute import SimpleImputer
    from sklearn.preprocessing import StandardScaler
    from sklearn.ensemble import RandomForestClassifier
    from sklearn.neural_network import MLPClassifier

    pipeline_choice = Choice(
        [SimpleImputer(), RandomForestClassifier()],
        [StandardScaler(), MLPClassifier()],
        name="pipeline_choice"
    )
    from amltk._doc import doc_print; doc_print(print, pipeline_choice)  # markdown-exec: hide
    ```

    Like all [`Node`][amltk.pipeline.node.Node]s, a `Sequential` accepts an explicit
    [`name=`][amltk.pipeline.node.Node.name],
    [`item=`][amltk.pipeline.node.Node.item],
    [`config=`][amltk.pipeline.node.Node.config],
    [`space=`][amltk.pipeline.node.Node.space],
    [`fidelities=`][amltk.pipeline.node.Node.fidelities],
    [`config_transform=`][amltk.pipeline.node.Node.config_transform] and
    [`meta=`][amltk.pipeline.node.Node.meta].

    See Also:
        * [`Node`][amltk.pipeline.node.Node]
    """  # noqa: E501

    nodes: tuple[Node, ...]
    """The nodes in series."""

    RICH_OPTIONS: ClassVar[RichOptions] = RichOptions(
        panel_color="#7E6B8F",
        node_orientation="vertical",
    )
    _NODES_INIT: ClassVar = "args"

    def __init__(
        self,
        *nodes: Node | NodeLike,
        name: str | None = None,
        item: Item | Callable[[Item], Item] | None = None,
        config: Config | None = None,
        space: Space | None = None,
        fidelities: Mapping[str, Any] | None = None,
        config_transform: Callable[[Config, Any], Config] | None = None,
        meta: Mapping[str, Any] | None = None,
    ):
        """See [`Node`][amltk.pipeline.node.Node] for details."""
        _nodes = tuple(as_node(n) for n in nodes)

        # Perhaps we need to do a deeper check on this...
        if not all_unique(_nodes, key=lambda node: node.name):
            raise DuplicateNamesError(self)

        if name is None:
            name = f"Seq-{randuid(8)}"

        super().__init__(
            *_nodes,
            name=name,
            item=item,
            config=config,
            space=space,
            fidelities=fidelities,
            config_transform=config_transform,
            meta=meta,
        )

    @property
    def tail(self) -> Node:
        """The last step in the pipeline."""
        return self.nodes[-1]

    def __len__(self) -> int:
        """Get the number of nodes in the pipeline."""
        return len(self.nodes)

    @override
    def __rshift__(self, other: Node | NodeLike) -> Sequential:
        other_node = as_node(other)
        if any(other_node.name == node.name for node in self.nodes):
            raise ValueError(
                f"Can't handle node with name '{other_node.name} as"
                f" there is already a node called '{other_node.name}' in {self.name}",
            )

        nodes = (*tuple(as_node(n) for n in self.nodes), other_node)
        return self.mutate(name=self.name, nodes=nodes)

    @override
    def walk(
        self,
        path: Sequence[Node] | None = None,
    ) -> Iterator[tuple[list[Node], Node]]:
        """Walk the nodes in this chain.

        Args:
            path: The current path to this node

        Yields:
            The parents of the node and the node itself
        """
        path = list(path) if path is not None else []
        yield path, self

        path = [*path, self]
        for node in self.nodes:
            yield from node.walk(path=path)

            # Append the previous node so that the next node in the sequence is
            # lead to from the previous node
            path = [*path, node]


@dataclass(init=False, frozen=True, eq=True)
class Split(Node[Item, Space]):
    """A [`Split`][amltk.pipeline.Split] of data in a pipeline.

    This indicates the different children in
    [`.nodes`][amltk.pipeline.Node.nodes] should
    act in parallel but on different subsets of data.

    ```python exec="true" source="material-block" html="true"
    from amltk.pipeline import Component, Split
    from sklearn.impute import SimpleImputer
    from sklearn.preprocessing import OneHotEncoder
    from sklearn.compose import make_column_selector

    categorical_pipeline = [
        SimpleImputer(strategy="constant", fill_value="missing"),
        OneHotEncoder(drop="first"),
    ]
    numerical_pipeline = Component(SimpleImputer, space={"strategy": ["mean", "median"]})

    preprocessor = Split(
        {
            "categories": categorical_pipeline,
            "numerical": numerical_pipeline,
        },
        config={
            # This is how you would configure the split for the sklearn builder in particular
            "categories": make_column_selector(dtype_include="category"),
            "numerical": make_column_selector(dtype_exclude="category"),
        },
        name="my_split"
    )
    from amltk._doc import doc_print; doc_print(print, preprocessor)  # markdown-exec: hide

    space = preprocessor.search_space("configspace")
    from amltk._doc import doc_print; doc_print(print, space)  # markdown-exec: hide

    configuration = space.sample_configuration()
    from amltk._doc import doc_print; doc_print(print, configuration)  # markdown-exec: hide

    configured_preprocessor = preprocessor.configure(configuration)
    from amltk._doc import doc_print; doc_print(print, configured_preprocessor)  # markdown-exec: hide

    built_preprocessor = configured_preprocessor.build("sklearn")
    print(built_preprocessor._repr_html_())  # markdown-exec: hide
    ```

    The split is a slight oddity when compared to the other kinds of components in that
    it allows a `dict` as it's first argument, where the keys are the names of the
    different paths through which data will go and the values are the actual nodes that
    will receive the data.

    If nodes are passed in as they are for all other components, usually the name of the
    first node will be important for any builder trying to make sense of how
    to use the `Split`


    Like all [`Node`][amltk.pipeline.node.Node]s, a `Split` accepts an explicit
    [`name=`][amltk.pipeline.node.Node.name],
    [`item=`][amltk.pipeline.node.Node.item],
    [`config=`][amltk.pipeline.node.Node.config],
    [`space=`][amltk.pipeline.node.Node.space],
    [`fidelities=`][amltk.pipeline.node.Node.fidelities],
    [`config_transform=`][amltk.pipeline.node.Node.config_transform] and
    [`meta=`][amltk.pipeline.node.Node.meta].

    See Also:
        * [`Node`][amltk.pipeline.node.Node]
    """  # noqa: E501

    nodes: tuple[Node, ...]
    """The nodes that this node leads to."""

    RICH_OPTIONS: ClassVar[RichOptions] = RichOptions(
        panel_color="#777DA7",
        node_orientation="horizontal",
    )

    _NODES_INIT: ClassVar = "args"

    def __init__(
        self,
        *nodes: Node | NodeLike | dict[str, Node | NodeLike],
        name: str | None = None,
        item: Item | Callable[[Item], Item] | None = None,
        config: Config | None = None,
        space: Space | None = None,
        fidelities: Mapping[str, Any] | None = None,
        config_transform: Callable[[Config, Any], Config] | None = None,
        meta: Mapping[str, Any] | None = None,
    ):
        """See [`Node`][amltk.pipeline.node.Node] for details."""
        if any(isinstance(n, dict) for n in nodes):
            if len(nodes) > 1:
                raise ValueError(
                    "Can't handle multiple nodes with a dictionary as a node.\n"
                    f"{nodes=}",
                )
            _node = nodes[0]
            assert isinstance(_node, dict)

            def _construct(key: str, value: Node | NodeLike) -> Node:
                match value:
                    case list():
                        return Sequential(*value, name=key)
                    case set() | tuple():
                        return as_node(value, name=key)
                    case _:
                        return Sequential(value, name=key)

            _nodes = tuple(_construct(key, value) for key, value in _node.items())
        else:
            _nodes = tuple(as_node(n) for n in nodes)

        if not all_unique(_nodes, key=lambda node: node.name):
            raise ValueError(
                f"Can't handle nodes they do not all contain unique names, {nodes=}."
                "\nAll nodes must have a unique name. Please provide a `name=` to them",
            )

        if name is None:
            name = f"Split-{randuid(8)}"

        super().__init__(
            *_nodes,
            name=name,
            item=item,
            config=config,
            space=space,
            fidelities=fidelities,
            config_transform=config_transform,
            meta=meta,
        )


@dataclass(init=False, frozen=True, eq=True)
class Component(Node[Item, Space]):
    """A [`Component`][amltk.pipeline.Component] of the pipeline with
    a possible item and **no children**.

    This is the basic building block of most pipelines, it accepts
    as it's [`item=`][amltk.pipeline.node.Node.item] some function that will be
    called with [`build_item()`][amltk.pipeline.components.Component.build_item] to
    build that one part of the pipeline.

    When [`build_item()`][amltk.pipeline.Component.build_item] is called,
    The [`.config`][amltk.pipeline.node.Node.config] on this node will be passed
    to the function to build the item.

    A common pattern is to use a [`Component`][amltk.pipeline.Component] to
    wrap a constructor, specifying the [`space=`][amltk.pipeline.node.Node.space]
    and [`config=`][amltk.pipeline.node.Node.config] to be used when building the
    item.

    ```python exec="true" source="material-block" html="true"
    from amltk.pipeline import Component
    from sklearn.ensemble import RandomForestClassifier

    rf = Component(
        RandomForestClassifier,
        config={"max_depth": 3},
        space={"n_estimators": (10, 100)}
    )
    from amltk._doc import doc_print; doc_print(print, rf)  # markdown-exec: hide

    config = {"n_estimators": 50}  # Sample from some space or something
    configured_rf = rf.configure(config)

    estimator = configured_rf.build_item()
    from amltk._doc import doc_print; doc_print(print, estimator)  # markdown-exec: hide
    ```

    Whenever some other node sees a function/constructor, i.e. `RandomForestClassifier`,
    this will automatically be converted into a `Component`.

    ```python exec="true" source="material-block" html="true"
    from amltk.pipeline import Sequential
    from sklearn.ensemble import RandomForestClassifier

    pipeline = Sequential(RandomForestClassifier, name="my_pipeline")
    from amltk._doc import doc_print; doc_print(print, pipeline)  # markdown-exec: hide
    ```

    The default `.name` of a component is the name of the class/function that it will
    use. You can explicitly set the `name=` if you want to when constructing the
    component.

    Like all [`Node`][amltk.pipeline.node.Node]s, a `Component` accepts an explicit
    [`name=`][amltk.pipeline.node.Node.name],
    [`item=`][amltk.pipeline.node.Node.item],
    [`config=`][amltk.pipeline.node.Node.config],
    [`space=`][amltk.pipeline.node.Node.space],
    [`fidelities=`][amltk.pipeline.node.Node.fidelities],
    [`config_transform=`][amltk.pipeline.node.Node.config_transform] and
    [`meta=`][amltk.pipeline.node.Node.meta].

    See Also:
        * [`Node`][amltk.pipeline.node.Node]
    """

    item: Callable[..., Item]
    """A node which constructs an item in the pipeline."""

    nodes: tuple[()]
    """A component has no children."""

    RICH_OPTIONS: ClassVar[RichOptions] = RichOptions(panel_color="#E6AF2E")

    _NODES_INIT: ClassVar = None

    def __init__(
        self,
        item: Callable[..., Item],
        *,
        name: str | None = None,
        config: Config | None = None,
        space: Space | None = None,
        fidelities: Mapping[str, Any] | None = None,
        config_transform: Callable[[Config, Any], Config] | None = None,
        meta: Mapping[str, Any] | None = None,
    ):
        """See [`Node`][amltk.pipeline.node.Node] for details."""
        super().__init__(
            name=name if name is not None else entity_name(item),
            item=item,
            config=config,
            space=space,
            fidelities=fidelities,
            config_transform=config_transform,
            meta=meta,
        )

    def build_item(self, **kwargs: Any) -> Item:
        """Build the item attached to this component.

        Args:
            **kwargs: Any additional arguments to pass to the item

        Returns:
            Item
                The built item
        """
        config = self.config or {}
        return self.item(**{**config, **kwargs})


@dataclass(init=False, frozen=True, eq=True)
class Searchable(Node[None, Space]):  # type: ignore
    """A [`Searchable`][amltk.pipeline.Searchable]
    node of the pipeline which just represents a search space, no item attached.

    While not usually applicable to pipelines you want to build, this component
    is useful for creating a search space, especially if the the real pipeline you
    want to optimize can not be built directly. For example, if you are optimize
    a script, you may wish to use a `Searchable` to represent the search space
    of that script.

    ```python exec="true" source="material-block" html="true"
    from amltk.pipeline import Searchable

    script_space = Searchable({"mode": ["orange", "blue", "red"], "n": (10, 100)})
    from amltk._doc import doc_print; doc_print(print, script_space)  # markdown-exec: hide
    ```

    A `Searchable` explicitly does not allow for `item=` to be set, nor can it have
    any children. A `Searchable` accepts an explicit
    [`name=`][amltk.pipeline.node.Node.name],
    [`config=`][amltk.pipeline.node.Node.config],
    [`space=`][amltk.pipeline.node.Node.space],
    [`fidelities=`][amltk.pipeline.node.Node.fidelities],
    [`config_transform=`][amltk.pipeline.node.Node.config_transform] and
    [`meta=`][amltk.pipeline.node.Node.meta].

    See Also:
        * [`Node`][amltk.pipeline.node.Node]
    """  # noqa: E501

    item: None = None
    """A searchable has no item."""

    nodes: tuple[()] = ()
    """A component has no children."""

    RICH_OPTIONS: ClassVar[RichOptions] = RichOptions(panel_color="light_steel_blue")

    _NODES_INIT: ClassVar = None

    def __init__(
        self,
        space: Space | None = None,
        *,
        name: str | None = None,
        config: Config | None = None,
        fidelities: Mapping[str, Any] | None = None,
        config_transform: Callable[[Config, Any], Config] | None = None,
        meta: Mapping[str, Any] | None = None,
    ):
        """See [`Node`][amltk.pipeline.node.Node] for details."""
        if name is None:
            name = f"Searchable-{randuid(8)}"

        super().__init__(
            name=name,
            config=config,
            space=space,
            fidelities=fidelities,
            config_transform=config_transform,
            meta=meta,
        )


@dataclass(init=False, frozen=True, eq=True)
class Fixed(Node[Item, None]):  # type: ignore
    """A [`Fixed`][amltk.pipeline.Fixed] part of the pipeline that
    represents something that can not be configured and used directly as is.

    It consists of an [`.item`][amltk.pipeline.node.Node.item] that is fixed,
    non-configurable and non-searchable. It also has no children.

    This is useful for representing parts of the pipeline that are fixed, for example
    if you have a pipeline that is a `Sequential` of nodes, but you want to
    fix the first component to be a `PCA` with `n_components=3`, you can use a `Fixed`
    to represent that.

    ```python exec="true" source="material-block" html="true"
    from amltk.pipeline import Component, Fixed, Sequential
    from sklearn.ensemble import RandomForestClassifier
    from sklearn.decomposition import PCA

    rf = Component(RandomForestClassifier, space={"n_estimators": (10, 100)})
    pca = Fixed(PCA(n_components=3))

    pipeline = Sequential(pca, rf, name="my_pipeline")
    from amltk._doc import doc_print; doc_print(print, pipeline)  # markdown-exec: hide
    ```

    Whenever some other node sees an instance of something, i.e. something that can't be
    called, this will automatically be converted into a `Fixed`.

    ```python exec="true" source="material-block" html="true"
    from amltk.pipeline import Sequential
    from sklearn.ensemble import RandomForestClassifier
    from sklearn.decomposition import PCA

    pipeline = Sequential(
        PCA(n_components=3),
        RandomForestClassifier(n_estimators=50),
        name="my_pipeline",
    )
    from amltk._doc import doc_print; doc_print(print, pipeline)  # markdown-exec: hide
    ```

    The default `.name` of a component is the class name of the item that it will
    use. You can explicitly set the `name=` if you want to when constructing the
    component.

    A `Fixed` accepts only an explicit [`name=`][amltk.pipeline.node.Node.name],
    [`item=`][amltk.pipeline.node.Node.item],
    [`meta=`][amltk.pipeline.node.Node.meta].

    See Also:
        * [`Node`][amltk.pipeline.node.Node]
    """

    item: Item = field()
    """The fixed item that this node represents."""

    space: None = None
    """A frozen node has no search space."""

    fidelities: None = None
    """A frozen node has no search space."""

    config: None = None
    """A frozen node has no config."""

    config_transform: None = None
    """A frozen node has no config so no transform."""

    nodes: tuple[()] = ()
    """A component has no children."""

    RICH_OPTIONS: ClassVar[RichOptions] = RichOptions(panel_color="#56351E")

    _NODES_INIT: ClassVar = None

    def __init__(
        self,
        item: Item,
        *,
        name: str | None = None,
        config: None = None,
        space: None = None,
        fidelities: None = None,
        config_transform: None = None,
        meta: Mapping[str, Any] | None = None,
    ):
        """See [`Node`][amltk.pipeline.node.Node] for details."""
        super().__init__(
            name=name if name is not None else entity_name(item),
            item=item,
            config=config,
            space=space,
            fidelities=fidelities,
            config_transform=config_transform,
            meta=meta,
        )



from __future__ import annotations

import os
from collections.abc import Callable
from functools import lru_cache
from typing import TYPE_CHECKING, Any, Literal

if TYPE_CHECKING:
    from rich.console import RenderableType

DEFAULT_MKDOCS_CODE_BLOCK_WIDTH = 80


SKLEARN_LINK = "https://www.scikit-learn.org/stable/modules/generated/{0}.html"


def sklearn_link_generator(name: str) -> str:
    """Generate a link for a sklearn function."""
    reduced_name = ".".join(s for s in name.split(".") if not s.startswith("_"))
    return SKLEARN_LINK.format(reduced_name)


@lru_cache
def _try_get_link(fully_scoped_name: str) -> str | None:
    """Try to get a link for a string.

    Expects fully qualified import names.
    """
    from amltk.options import _amltk_options

    links = _amltk_options.get("links", {})

    for k, v in links.items():
        if fully_scoped_name.startswith(k):
            if isinstance(v, str):
                return v
            if callable(v):
                return v(fully_scoped_name)

    return None


def link(obj: Any) -> str | None:
    """Try to get a link for an object."""
    from amltk._functional import fullname

    return _try_get_link(fullname(obj))


def make_picklable(thing: Any, name: str | None = None) -> None:
    """This is hack to make the examples code with schedulers work.

    Scheduler uses multiprocessing and multiprocessing requires that
    all objects passed to the scheduler are picklable. This is not
    the case for the classes/functions defined in the example code.
    """
    import __main__

    _name = thing.__name__ if name is None else name
    setattr(__main__, _name, thing)


def as_rich_svg(
    *renderable: RenderableType,
    title: str = "",
    width: int = DEFAULT_MKDOCS_CODE_BLOCK_WIDTH,
) -> str:
    from rich.console import Console

    with open(os.devnull, "w") as devnull:  # noqa: PTH123
        console = Console(record=True, width=width, file=devnull, markup=False)
        for r in renderable:
            console.print(r, markup=False)

    return console.export_svg(title=title)


HTML_FORMAT = """
<pre style="font-family:Menlo,'DejaVu Sans Mono',consolas,'Courier New',monospace;font-size:{fontsize}">
<code style="font-family:inherit">{{code}}</code>
</pre>
"""  # noqa: E501

SIZES = {
    "very-small": "0.5rem",
    "small": "0.7rem",
    "medium": "0.75rem",
    "large": "1rem",
}


def as_rich_html(
    *renderable: RenderableType,
    width: int = DEFAULT_MKDOCS_CODE_BLOCK_WIDTH,
    fontsize: str = "medium",
) -> str:
    from rich.console import Console

    with open(os.devnull, "w") as devnull:  # noqa: PTH123
        console = Console(record=True, width=width, file=devnull)
        for r in renderable:
            console.print(r, markup=False, highlight=False)

        fontsize = SIZES.get(fontsize, fontsize)
        code_format = HTML_FORMAT.format(fontsize=fontsize)
        return console.export_html(inline_styles=True, code_format=code_format)


def doc_print(
    _print: Callable[[str], Any],
    *renderable: RenderableType,
    title: str = "",
    width: int = DEFAULT_MKDOCS_CODE_BLOCK_WIDTH,
    output: Literal["svg", "html"] = "html",
    fontsize: str = "medium",
) -> None:
    if output == "svg":
        _print(as_rich_svg(*renderable, title=title, width=width))
    elif len(renderable) == 1:
        try:
            from sklearn.base import BaseEstimator, TransformerMixin
            from sklearn.pipeline import Pipeline

            if isinstance(
                renderable[0],
                Pipeline | BaseEstimator() | TransformerMixin(),
            ):
                _print(renderable[0]._repr_html_())  # type: ignore
                return
        except Exception:  # noqa: BLE001
            _print(as_rich_html(*renderable, width=width, fontsize=fontsize))
    else:
        _print(as_rich_html(*renderable, width=width, fontsize=fontsize))



"""[Optuna](https://optuna.org/) is an automatic hyperparameter optimization
software framework, particularly designed for machine learning.

!!! tip "Requirements"

    This requires `Optuna` which can be installed with:

    ```bash
    pip install amltk[optuna]

    # Or directly
    pip install optuna
    ```

We provide a thin wrapper called
[`OptunaOptimizer`][amltk.optimization.optimizers.optuna.OptunaOptimizer] from which
you can integrate `Optuna` into your workflow.

This uses an Optuna-like [`search_space()`][amltk.pipeline.Node.search_space] for
its optimization.

Users should report results using
[`trial.success()`][amltk.optimization.Trial.success]
with either `cost=` or `values=` depending on any optimization directions
given to the underyling optimizer created. Please see their documentation
for more.

Visit their documentation for what you can pass to
[`OptunaOptimizer.create()`][amltk.optimization.optimizers.optuna.OptunaOptimizer.create],
which is forward to [`optun.create_study()`][optuna.create_study].

```python exec="True" source="material-block" result="python"
from __future__ import annotations

import logging

from sklearn.datasets import load_iris
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import accuracy_score
from sklearn.model_selection import train_test_split

from amltk.optimization.optimizers.optuna import OptunaOptimizer
from amltk.scheduling import Scheduler
from amltk.optimization import History, Trial, Metric
from amltk.pipeline import Component

logging.basicConfig(level=logging.INFO)


def target_function(trial: Trial, pipeline: Pipeline) -> Trial.Report:
    X, y = load_iris(return_X_y=True)
    X_train, X_test, y_train, y_test = train_test_split(X, y)
    clf = pipeline.configure(trial.config).build("sklearn")

    with trial.begin():
        clf.fit(X_train, y_train)
        y_pred = clf.predict(X_test)
        return trial.success(accuracy=accuracy_score(y_test, y_pred))

    return trial.fail()
from amltk._doc import make_picklable; make_picklable(target_function)  # markdown-exec: hide


accuracy_metric = Metric("accuracy", minimize=False, bounds=(0, 1))
pipeline = Component(RandomForestClassifier, space={"n_estimators": (10, 100)})
optimizer = OptunaOptimizer.create(space=pipeline, metrics=accuracy_metric)

N_WORKERS = 2
scheduler = Scheduler.with_processes(N_WORKERS)
task = scheduler.task(target_function)

history = History()

@scheduler.on_start(repeat=N_WORKERS)
def on_start():
    trial = optimizer.ask()
    task.submit(trial, pipeline)

@task.on_result
def tell_and_launch_trial(_, report: Trial.Report):
    if scheduler.running():
        optimizer.tell(report)
        trial = optimizer.ask()
        task.submit(trial, pipeline)


@task.on_result
def add_to_history(_, report: Trial.Report):
    history.add(report)

scheduler.run(timeout=3, wait=False)

print(history.df())
optimizer.bucket.rmdir()  # markdown-exec: hide
```

!!! todo "Some more documentation"

    Sorry!

"""  # noqa: E501
from __future__ import annotations

from collections.abc import Sequence
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any
from typing_extensions import Self, override

import optuna
from more_itertools import first_true
from optuna.samplers import BaseSampler, NSGAIISampler, TPESampler
from optuna.study import Study, StudyDirection
from optuna.trial import (
    Trial as OptunaTrial,
    TrialState,
)

import amltk.randomness
from amltk.optimization import Optimizer, Trial
from amltk.optimization.metric import Metric
from amltk.pipeline import Node
from amltk.pipeline.parsers.optuna import parser
from amltk.store import PathBucket

if TYPE_CHECKING:
    from typing import Protocol

    from amltk.pipeline.parsers.optuna import OptunaSearchSpace
    from amltk.types import Seed

    class OptunaParser(Protocol):
        """A protocol for Optuna search space parser."""

        def __call__(
            self,
            node: Node,
            *,
            flat: bool = False,
            delim: str = ":",
        ) -> OptunaSearchSpace:
            """See [`optuna_parser`][amltk.pipeline.parsers.optuna.parser]."""
            ...


class OptunaOptimizer(Optimizer[OptunaTrial]):
    """An optimizer that uses Optuna to optimize a search space."""

    @override
    def __init__(
        self,
        *,
        study: Study,
        metrics: Metric | Sequence[Metric],
        bucket: PathBucket | None = None,
        seed: Seed | None = None,
        space: OptunaSearchSpace,
    ) -> None:
        """Initialize the optimizer.

        Args:
            study: The Optuna Study to use.
            metrics: The metrics to optimize.
            bucket: The bucket given to trials generated by this optimizer.
            space: Defines the current search space.
            seed: The seed to use for the sampler and trials.
        """
        # Verify the study has the same directions as the metrics
        match metrics:
            case Metric(minimize=minimize):
                _dir = StudyDirection.MINIMIZE if minimize else StudyDirection.MAXIMIZE
                if study.direction != _dir:
                    raise ValueError(
                        f"The study direction is {_dir}, but the metric minimize is "
                        f"{minimize}.",
                    )
            case metrics:
                _dirs = [
                    StudyDirection.MINIMIZE if m.minimize else StudyDirection.MAXIMIZE
                    for m in metrics
                ]
                if study.directions != _dirs:
                    raise ValueError(
                        f"The study directions are {_dirs}, but the metrics minimize "
                        f"are {[m.minimize for m in metrics]}.",
                    )

        metrics = [metrics] if isinstance(metrics, Metric) else metrics
        super().__init__(bucket=bucket, metrics=metrics)
        self.seed = amltk.randomness.as_int(seed)
        self.study = study
        self.metrics = metrics
        self.space = space

    @classmethod
    def create(
        cls,
        *,
        space: OptunaSearchSpace | Node,
        metrics: Metric | Sequence[Metric],
        bucket: PathBucket | str | Path | None = None,
        sampler: BaseSampler | None = None,
        seed: Seed | None = None,
        **kwargs: Any,
    ) -> Self:
        """Create a new Optuna optimizer. For more information, check Optuna
            documentation
            [here](https://optuna.readthedocs.io/en/stable/reference/generated/optuna.study.create_study.html#).

        Args:
            space: Defines the current search space.
            metrics: The metrics to optimize.
            bucket: The bucket given to trials generated by this optimizer.
            sampler: The sampler to use. Default is to use:

                * Single metric: [TPESampler][optuna.samplers.TPESampler]
                * Multiple metrics: [NSGAIISampler][optuna.samplers.NSGAIISampler]

            seed: The seed to use for the sampler and trials.

            **kwargs: Additional arguments to pass to
                [`optuna.create_study`][optuna.create_study].

        Returns:
            Self: The newly created optimizer.
        """
        if "direction" in kwargs:
            raise ValueError(
                "The direction should be provided through the 'metrics' argument.",
            )

        if isinstance(space, Node):
            space = space.search_space(parser=cls.preferred_parser())

        match bucket:
            case None:
                bucket = PathBucket(
                    f"{cls.__class__.__name__}-{datetime.now().isoformat()}",
                )
            case str() | Path():
                bucket = PathBucket(bucket)
            case bucket:
                bucket = bucket  # noqa: PLW0127

        match metrics:
            case Metric(minimize=minimize):
                direction = (
                    StudyDirection.MINIMIZE if minimize else StudyDirection.MAXIMIZE
                )
                study = optuna.create_study(direction=direction, **kwargs)
            case metrics:
                directions = [
                    StudyDirection.MINIMIZE if m.minimize else StudyDirection.MAXIMIZE
                    for m in metrics
                ]
                study = optuna.create_study(directions=directions, **kwargs)

        if sampler is None:
            sampler_seed = amltk.randomness.as_int(seed)
            match metrics:
                case Metric():
                    sampler = TPESampler(seed=sampler_seed)  # from `create_study()`
                case metrics:
                    sampler = NSGAIISampler(seed=sampler_seed)  # from `create_study()`

        return cls(study=study, metrics=metrics, space=space, bucket=bucket, seed=seed)

    @override
    def ask(self) -> Trial[OptunaTrial]:
        """Ask the optimizer for a new config.

        Returns:
            The trial info for the new config.
        """
        optuna_trial: optuna.Trial = self.study.ask(self.space)
        config = optuna_trial.params
        trial_number = optuna_trial.number
        unique_name = f"{trial_number=}"
        metrics = [self.metrics] if isinstance(self.metrics, Metric) else self.metrics
        return Trial(
            name=unique_name,
            seed=self.seed,
            config=config,
            info=optuna_trial,
            bucket=self.bucket,
            metrics=metrics,
        )

    @override
    def tell(self, report: Trial.Report[OptunaTrial]) -> None:
        """Tell the optimizer the result of the sampled config.

        Args:
            report: The report of the trial.
        """
        trial = report.trial.info
        assert trial is not None

        match report.status:
            case Trial.Status.CRASHED | Trial.Status.UNKNOWN | Trial.Status.FAIL:
                # NOTE: Can't tell any values if the trial crashed or failed
                self.study.tell(trial=trial, state=TrialState.FAIL)
            case Trial.Status.SUCCESS:
                match self.metrics:
                    case [metric]:
                        metric_value: Metric.Value = first_true(
                            report.metric_values,
                            pred=lambda m: m.metric == metric,
                            default=metric.worst,
                        )
                        self.study.tell(
                            trial=trial,
                            state=TrialState.COMPLETE,
                            values=metric_value.value,
                        )
                    case metrics:
                        # NOTE: We need to make sure that there sorted in the order
                        # that Optuna expects, with any missing metrics filled in
                        _lookup = {v.metric.name: v for v in report.metric_values}
                        values = [
                            _lookup.get(metric.name, metric.worst).value
                            for metric in metrics
                        ]
                        self.study.tell(
                            trial=trial,
                            state=TrialState.COMPLETE,
                            values=values,
                        )

    @override
    @classmethod
    def preferred_parser(cls) -> OptunaParser:
        return parser



"""Get an XGBoost component for your pipeline.

A [`Component`][amltk.pipeline.Component] wrapped
around [XGBoost](https://xgboost.readthedocs.io/en/stable/index.html) with
two possible default spaces `#!python "large"` and `#!python "performant"`
or you own custom `space=`.

See [`amltk.pipeline.xgboost.xgboost_component`][amltk.pipeline.xgboost.xgboost_component]
"""  # noqa: E501


from __future__ import annotations

import warnings
from collections.abc import Mapping
from typing import Any, Literal

from xgboost import XGBClassifier, XGBRegressor

from amltk.pipeline import Component


def xgboost_component(
    kind: Literal["classifier", "regressor"],
    name: str | None = None,
    space: Any | Literal["large", "performant"] = "large",
    config: Mapping[str, Any] | None = None,
) -> Component:
    """Create a component with an XGBoost estimator.

    Args:
        kind: The kind of estimator to create, either "classifier" or "regressor".
        name: The name of the step in the pipeline.
        space: The space to use for hyperparameter optimization. Choose from
            "large", "performant" or provide a custom space.

            !!! todo

                Currently `#!python "performant"` links to `#!python "large"` and are
                the same.

            !!! warning "Warning"

                The default space is by no means optimal, please adjust it to your
                needs. You can find the link to the XGBoost parameters here:

                https://xgboost.readthedocs.io/en/stable/parameter.html


        config: The keyword arguments to pass to the XGBoost estimator when it will be
            created. These will be hard set on the estimator and removed from the
            default space if no space is provided.

    Returns:
        A pipeline with an XGBoost estimator.
    """
    estimator_types = {
        "classifier": XGBClassifier,
        "regressor": XGBRegressor,
    }
    config = config or {}
    estimator_type = estimator_types.get(kind)
    if estimator_type is None:
        raise ValueError(
            f"Unknown kind: {kind}, please choose from {list(estimator_types.keys())}",
        )

    if name is None:
        name = str(estimator_type.__name__)

    if isinstance(space, str):
        device = config.get("device", "cpu")
        tree_method = config.get("tree_method", "auto")
        is_classifier = estimator_type is XGBClassifier
        _spaces = {
            "large": xgboost_large_space,
            "performant": xgboost_performant_space,
        }
        space_f = _spaces[space]
        space = space_f(
            is_classifier=is_classifier,
            tree_method=tree_method,
            device=device,
        )
        for key in config:
            space.pop(key, None)
    elif isinstance(space, Mapping):
        overlap = set(space.keys()).intersection(config)
        if any(overlap):
            raise ValueError(
                f"Space and kwargs overlap: {overlap}, please remove one of them",
            )

    return Component(name=name, item=estimator_type, config=config, space=space)


def xgboost_large_space(
    *,
    is_classifier: bool,
    tree_method: str,
    device: str,
) -> dict[str, Any]:
    """Create a large space for XGBoost hyperparameter optimization."""
    # TODO: Do we want a general space kind
    # For now we use ConfigSpace where needed
    from ConfigSpace import Float, Integer

    space = {
        "eta": Float("eta", (1e-3, 1), default=0.3, log=True),
        "min_split_loss": Float("min_split_loss", (0, 20), default=0),
        "max_depth": Integer("max_depth", (1, 20), default=6),
        "min_child_weight": Float("min_child_weight", (0, 20), default=1),
        "max_delta_step": Float("max_delta_step", (0, 10), default=0),
        "subsample": Float("subsample", (1e-5, 1), default=1),
        "colsample_bytree": Float("colsample_bytree", (1e-5, 1), default=1),
        "colsample_bylevel": Float("colsample_bylevel", (1e-5, 1), default=1),
        "colsample_bynode": Float("colsample_bynode", (1e-5, 1), default=1),
        "reg_lambda": Float("reg_lambda", (1e-5, 1e3), default=1),
        "reg_alpha": Float("reg_alpha", (0, 1e3), default=0),
    }

    if tree_method == "hist" and ("cuda" in device or "gpu" in device):
        space["sampling_method"] = ["uniform", "gradient_based"]

    if tree_method in ("hist", "approx"):
        space["max_bin"] = Integer("max_bin", (2, 512), default=256)

    if is_classifier:
        space["scale_pos_weight"] = Float("scale_pos_weight", (1e-5, 1e5), default=1)

    return space


def xgboost_performant_space(
    *,
    is_classifier: bool,
    tree_method: str,
    device: str,
) -> dict[str, Any]:
    """Create a performant space for XGBoost hyperparameter optimization."""
    warnings.warn(
        "This space is not yet optimized for performance"
        " and is subject to change in the future.",
        FutureWarning,
        stacklevel=2,
    )
    return xgboost_large_space(
        is_classifier=is_classifier,
        tree_method=tree_method,
        device=device,
    )



"""A `Drop` in a [`Bucket`][amltk.store.bucket.Bucket]
is a reference to a resource.
"""
from __future__ import annotations

import logging
from collections.abc import Callable
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Generic, TypeVar, overload

from more_itertools.more import first

from amltk._functional import funcname
from amltk.types import StoredValue

if TYPE_CHECKING:
    from amltk.store.loader import Loader


logger = logging.getLogger(__name__)

T = TypeVar("T")
Default = TypeVar("Default")
KeyT = TypeVar("KeyT")


@dataclass
class Drop(Generic[KeyT]):
    """A drop is a reference to a resource in a bucket.

    You likely do not need to create these yourself and
    are a class used by [`Bucket`][amltk.store.bucket.Bucket] to wrap
    access to a resource located at a give `key`.

    The main use of this class is to attempt to use different
    [`loaders`][amltk.store.loader.Loader] to load a resource
    at a given `key`, using the `key` to try infer which loader
    to use. Each drop has a list of default loaders that it will
    try to use to load the resource.

    For flexibility, you can also specify a `how` when using any
    of [`load`][amltk.store.drop.Drop.load], [`get`][amltk.store.drop.Drop.get]
    or [`put`][amltk.store.drop.Drop.put] to override the default loaders.
    The [`remove`][amltk.store.drop.Drop.remove] and
    [`exists`][amltk.store.drop.Drop.exists] method also has a `how`
    incase the default methods are not sufficient.

    To support well typed code, you can also specify a `check` type
    which will be used to checked when loading objects, to make sure
    it is of the correct type. This is ignored if `how` is specified.


    The primary methods of interest are
    * [`load`][amltk.store.drop.Drop.load]
    * [`get`][amltk.store.drop.Drop.get]
    * [`put`][amltk.store.drop.Drop.put]
    * [`remove`][amltk.store.drop.Drop.remove]
    * [`exists`][amltk.store.drop.Drop.exists]
    * [`as_stored_value`][amltk.store.drop.Drop.as_stored_value]

    Args:
        key: The key to the resource.
        loaders: The loaders to use to load the resource.
    """

    key: KeyT
    loaders: tuple[type[Loader[KeyT, Any]], ...] = field(repr=False)
    _remove: Callable[[KeyT], bool] = field(repr=False)
    _exists: Callable[[KeyT], bool] = field(repr=False)

    def as_stored_value(
        self,
        read: Callable[[KeyT], T] | None = None,
    ) -> StoredValue[KeyT, T]:
        """Convert the drop to a [`StoredValue`][amltk.types.StoredValue].

        Args:
            read: The method to use to load the resource. If `None` then
                the first loader that can load the resource will be used.

        Returns:
            The drop as a [`StoredValue`][amltk.types.StoredValue].
        """
        if read is None:
            loader = first(
                (_l for _l in self.loaders if _l.can_load(self.key)),
                default=None,
            )

            if loader is None:
                raise ValueError(f"Can't load {self.key=} from {self.loaders=}")

            read = loader.load

        return StoredValue(self.key, read=read)

    def put(
        self,
        obj: T,
        *,
        how: Callable[[T], None] | None = None,
    ) -> None:
        """Put an object into the bucket.

        Args:
            obj: The object to put into the bucket.
            how: The function to use to put the object into the bucket.
                If `None` then the first loader that can put the object
                will be used.
        """
        if how:
            how(obj)
            return

        loader = first(
            (_l for _l in self.loaders if _l.can_save(obj, self.key)),
            default=None,
        )
        if not loader:
            msg = (
                f"No default way to handle {type(obj)=} objects."
                " Please provide a `how` function that will save"
                f" the object to {self.key}."
            )
            raise ValueError(msg)

        loader.save(obj, self.key)

    @overload
    def load(self, *, check: None = None, how: None = None) -> Any:
        ...

    @overload
    def load(self, *, check: type[T], how: None = None) -> T:
        ...

    @overload
    def load(self, *, check: type[T] | None = ..., how: Callable[[KeyT], T]) -> T:
        ...

    def load(
        self,
        *,
        check: type[T] | None = None,
        how: Callable[[KeyT], T] | None = None,
    ) -> T | Any:
        """Load the resource.

        Args:
            check: By specifying a `type` we check the loaded object of that type, to
                enable correctly typed checked code.
            how: The function to use to load the resource.

        Returns:
            The loaded resource.
        """
        if not isinstance(how, type) and callable(how):
            value = how(self.key)
            loader_name = funcname(how)
        else:
            loader = first(
                (_l for _l in self.loaders if _l.can_load(self.key)),
                default=None,
            )
            if loader is None:
                raise ValueError(f"Can't load {self.key=} from {self.loaders=}")

            value = loader.load(self.key)
            loader_name = loader.name

        if check is not None and not isinstance(value, check):
            msg = (
                f"Value {value=} loaded by {loader_name=} is not of type {check=},"
                f" but is of type {type(value)=}."
            )
            raise TypeError(msg)

        return value

    @overload
    def get(
        self,
        default: None = None,
        *,
        check: None = None,
        how: None = None,
    ) -> Any | None:
        ...

    @overload
    def get(
        self,
        default: Default,
        *,
        check: None = None,
        how: None = None,
    ) -> Default | None:
        ...

    @overload
    def get(
        self,
        default: None = None,
        *,
        check: type[T],
        how: Callable[[KeyT], T] = ...,
    ) -> T | None:
        ...

    @overload
    def get(
        self,
        default: Default,
        *,
        check: type[T],
        how: Callable[[KeyT], T],
    ) -> Default | T:
        ...

    @overload
    def get(
        self,
        default: Default,
        *,
        check: type[T],
        how: Callable[[KeyT], T] | None = ...,
    ) -> Default | T:
        ...

    def get(
        self,
        default: Default | None = None,
        *,
        check: type[T] | None = None,
        how: Callable[[KeyT], T] | None = None,
    ) -> Default | T | None:
        """Load the resource, or return the default if it can't be loaded.

        See [`load`][amltk.store.drop.Drop.load] for more details.

        Note:
            This function makes no distinction for the reason it fails to load,
            namely if it's of the incorrect type or the resource does not exist
            at the key.

        Args:
            default: The default value to return if the resource can't be loaded.
            check: By specifying a `type` we check the loaded object of that type, to
                enable correctly typed checked code. If the default value should
                be returned because the resource can't be loaded, then the default
                value is **not** checked.
            how: The function to use to load the resource.

        Returns:
            The loaded resource or the default value if it cant be loaded.
        """
        try:
            return self.load(check=check, how=how)
        except TypeError as e:
            raise e
        except FileNotFoundError:
            return default
        except Exception as e:  # noqa: BLE001
            logger.warning(
                f"Failed to load {self.key=} from {self.loaders=}: {e}",
                exc_info=True,
            )

        return None

    def remove(self, *, how: Callable[[KeyT], bool] | None = None) -> bool:
        """Remove the resource from the bucket.

        Args:
            how: The function to use to remove the resource. Returns `True` if
                the resource no longer exists after the removal, `False` otherwise.

                !!! note "Non-existent resources"

                    If the resource does not exist, then the function will `True`.
        """
        logger.debug(f"Removing {self.key=}")
        if how:
            return how(self.key)

        return self._remove(self.key)

    def exists(self, *, how: Callable[[KeyT], bool] | None = None) -> bool:
        """Check if the resource exists.

        Returns:
            `True` if the resource exists, `False` otherwise.
        """
        if how:
            return how(self.key)
        return self._exists(self.key)



"""A module containing a concreate implementation of a
[`Bucket`][amltk.store.bucket.Bucket] that uses the Path API to store objects.
"""
from __future__ import annotations

import shutil
from collections.abc import Iterator, Sequence
from itertools import chain
from pathlib import Path
from typing import TYPE_CHECKING, Any
from typing_extensions import override

from amltk.store.bucket import Bucket
from amltk.store.drop import Drop
from amltk.store.paths.path_loaders import (
    ByteLoader,
    JSONLoader,
    NPYLoader,
    PathLoader,
    PDLoader,
    PickleLoader,
    TxtLoader,
    YAMLLoader,
)

if TYPE_CHECKING:
    from typing_extensions import Self


DEFAULT_LOADERS: tuple[type[PathLoader], ...] = (
    NPYLoader,
    PDLoader,
    JSONLoader,  # We prefer json over yaml
    YAMLLoader,
    TxtLoader,
    ByteLoader,
    PickleLoader,
)


class PathBucket(Bucket[str, Path]):
    """A bucket that uses the Path API to store objects.

    This bucket is a key-value lookup backed up by some filesystem.
    By assinging to the bucket, you store the object to the filesystem.
    However the values you get back are instead a [`Drop`][amltk.store.drop.Drop]
    that can be used to perform operations on the stores object, such as `load`, `get`
    and `remove`.

    ???+ note "Drop methods"
        * [`Drop.load`][amltk.store.drop.Drop.load] - Load the object from the bucket.
        * [`Drop.get`][amltk.store.drop.Drop.get] - Load the object from the bucket
            with a default if something fails.
        * [`Drop.put`][amltk.store.drop.Drop.put] - Store an object in the bucket.
        * [`Drop.remove`][amltk.store.drop.Drop.remove] - Remove the object from the
            bucket.
        * [`Drop.exists`][amltk.store.drop.Drop.exists] - Check if the object exists
            in the bucket.

    ```python
    from amltk.store.paths import PathBucket
    import numpy as np
    import pandas as pd
    from sklearn.linear_model import LinearRegression

    bucket = PathBucket("path/to/bucket")

    array = np.array([1, 2, 3])
    dataframe = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    model = LinearRegression()

    # Store things
    bucket["myarray.npy"] = array # (1)!
    bucket["df.csv"] = dataframe  # (2)!
    bucket["model.pkl"].put(model)

    bucket["config.json"] = {"hello": "world"}
    assert bucket["config.json"].exists()
    bucket["config.json"].remove()

    # Load things
    array = bucket["myarray.npy"].load()
    maybe_df = bucket["df.csv"].get()  # (3)!
    model: LinearRegression = bucket["model.pkl"].get(check=LinearRegression)  # (4)!

    # Create subdirectories
    model_bucket = bucket / "my_model" # (5)!
    model_bucket["model.pkl"] = model
    model_bucket["predictions.npy"] = model.predict(X)

    # Acts like a mapping
    assert "myarray.npy" in bucket
    assert len(bucket) == 3
    for key, item in bucket.items():
        print(key, item.load())
    del bucket["model.pkl"]
    ```

    1. The `=` is a shortcut for `bucket["myarray.npy"].put(array)`
    2. The extension is used to determine which
        [`PathLoader`][amltk.store.paths.path_loaders.PathLoader] to use
        and how to save it.
    3. The `get` method acts like the [`dict.load`][dict] method.
    4. The `get` method can be used to check the type of the loaded object.
        If the type does not match, a `TypeError` is raised.
    5. Uses the familiar [`Path`][pathlib.Path] API to create subdirectories.
    """

    def __init__(
        self,
        path: Path | str,
        *,
        loaders: Sequence[type[PathLoader]] | None = None,
        create: bool = True,
        clean: bool = False,
        exists_ok: bool = True,
    ) -> None:
        """Create a new PathBucket.

        Args:
            path: The path to the bucket.
            loaders: A sequence of loaders to use when loading objects.
                These will be prepended to the default loaders and attempted
                to be used first.
            create: If True, the base path will be created if it does not
                exist.
            clean: If True, the base path will be deleted if it exists.
            exists_ok: If False, an error will be raised if the base path
                already exists.
        """
        super().__init__()
        _loaders = DEFAULT_LOADERS
        if loaders is not None:
            _loaders = tuple(chain(loaders, DEFAULT_LOADERS))

        if isinstance(path, str):
            path = Path(path)

        if clean and path.exists():
            shutil.rmtree(path, ignore_errors=True)

        if not exists_ok and path.exists():
            raise FileExistsError(f"File/Directory already exists at {path}")

        if create:
            path.mkdir(parents=True, exist_ok=True)

        self._create = create
        self.path = path
        self.loaders = _loaders

    def sizes(self) -> dict[str, int]:
        """Get the sizes of all the files in the bucket.

        !!! warning "Files only"

            This method only returns the sizes of the files in the bucket.
            It does not include directories, their sizes, or their contents.

        Returns:
            A dictionary mapping the keys to the sizes of the files.
        """
        return {str(path.name): path.stat().st_size for path in self.path.iterdir()}

    def add_loader(self, loader: type[PathLoader]) -> None:
        """Add a loader to the bucket.

        Args:
            loader: The loader to add.
        """
        self.loaders = (loader, *self.loaders)

    @override
    def __getitem__(self, key: str) -> Drop[Path]:
        return self._drop(self.path / key, loaders=self.loaders)

    @override
    def __setitem__(self, key: str, value: Any) -> None:
        self._drop(self.path / key, loaders=self.loaders).put(value)

    @override
    def __delitem__(self, key: str) -> None:
        self._drop(self.path / key, loaders=self.loaders).remove()

    @override
    def __iter__(self) -> Iterator[str]:
        return (path.name for path in self.path.iterdir())

    @override
    def sub(self, key: str, *, create: bool | None = None) -> Self:
        """Create a subdirectory of the bucket.

        Args:
            key: The name of the subdirectory.
            create: Whether the subdirectory will be created if it does not
                exist. If None, the default, the value of `create` passed to
                the constructor will be used.

        Returns:
            A new bucket with the same loaders as the current bucket.
        """
        return self.__class__(
            self.path / key,
            loaders=self.loaders,
            create=self._create if create is None else create,
            clean=False,
        )

    @override
    def __contains__(self, key: object) -> bool:
        if not isinstance(key, str):
            return False
        return (self.path / key).exists()

    @classmethod
    def _drop(cls, path: Path, loaders: tuple[type[PathLoader], ...]) -> Drop[Path]:
        return Drop(
            path,
            loaders=loaders,
            _remove=cls._remove,
            _exists=cls._exists,
        )

    @classmethod
    def _remove(cls, path: Path) -> bool:
        if not path.exists():
            return True

        try:
            if path.is_dir():
                shutil.rmtree(path)
            else:
                path.unlink()

            return True
        except OSError:
            return False

    @classmethod
    def _exists(cls, path: Path) -> bool:
        return path.exists()

    @override
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.path!r})"

    def rmdir(self) -> None:
        """Delete the bucket."""
        shutil.rmtree(self.path)



from amltk._richutil.renderable import RichRenderable
from amltk._richutil.renderers import Function, rich_make_column_selector
from amltk._richutil.util import df_to_table, richify

__all__ = [
    "df_to_table",
    "richify",
    "RichRenderable",
    "Function",
    "rich_make_column_selector",
]



"""Distance functions.

This module contains functions for calculating the distance between
two vectors.
"""
from __future__ import annotations

from collections.abc import Callable
from functools import partial
from typing import Any, Literal, TypeAlias

import numpy as np
import numpy.typing as npt

_norm = np.linalg.norm

DistanceMetric: TypeAlias = Callable[[npt.ArrayLike, npt.ArrayLike], float]
"""A metric used for calculating distances.

Takes two arrays-like objects and returns a float.
"""


def pnorm(
    x: npt.ArrayLike,
    y: npt.ArrayLike,
    p: int | float = 2,
) -> float:
    """Calculates the p-norm between each column in x and y.

    The p-norm is defined as:

        `||x - y||_p = (sum_i(|x_i - y_i|^p))^(1/p)`

    The common values for p are 1, 2 and infinity.

    * [`l1_distance()`][amltk.distances.l1_distance]
    * [`l2_distance()`][amltk.distances.l2_distance]
    * [`linf_distance()`][amltk.distances.linf_distance]

    !!! tip "Using a `partial`"

        To use this function with
        [`dataset_distance()`][amltk.metalearning.dataset_distance],
        you can wrap this in [`functools.partial()`][functools.partial].

        ```python
        from functools import partial
        from amltk.metalearning import dataset_distance
        from amltk.distances import pnorm

        dataset_distance(
            target,
            dataset_metafeatures,
            method=partial(pnorm, p=3), # (1)!
        )
        ```

        1. [`partial()`][functools.partial] creates a new function with the
        `p` argument set to 3.

    Args:
        x: The vector to compare.
        y: The vector to compute the distance to
        p: The p in p-norm.

    Returns:
        A series with the same index as x.
    """
    x = np.asarray(x)
    y = np.asarray(y)

    if p is np.inf:
        return float(np.max(np.abs(x - y)))

    return float(np.linalg.norm(x - y, ord=p))


def cosine_distance(x: npt.ArrayLike, y: npt.ArrayLike) -> float:
    """Calculates the cosine distance between each column in x and y.

    The cosine distance is defined as 1 - cosine_similarity. This means
    the distance is 0 when the vectors are identical, 1 when orthogonal
    and 2 when they are opposite.

    Args:
        x: A dataframe with columns being the features and rows being the samples.
        y: A series with the same index as x.

    Returns:
        A series with the same index as x.
    """
    x = np.asarray(x)
    y = np.asarray(y)

    cosine_similarity = np.dot(x, y) / (_norm(x) * _norm(y))
    return float(1 - cosine_similarity)


l1_distance = partial(pnorm, p=1)
"""Calculates the l1 distance between each column in x and y.

The l1 distance is defined as:

    `||x - y||_1 = sum_i(|x_i - y_i|)`

This is the sum of the absolute differences between each element in x and y.

See Also:
    * [`pnorm()`][amltk.distances.pnorm]
"""

l2_distance = partial(pnorm, p=2)
"""Calculates the l2 distance between each column in x and y.

The l2 distance is defined as:

    `||x - y||_2 = sqrt(sum_i(|x_i - y_i|^2))`

This is the square root of the sum of the squared differences between each
element in x and y.

See Also:
    * [`pnorm()`][amltk.distances.pnorm]
"""

linf_distance = partial(pnorm, p=np.inf)
"""Calculates the linf distance between each column in x and y.

The linf distance is defined as:

    `||x - y||_inf = max_i(|x_i - y_i|)`

This is the maximum absolute difference between each element in x and y.

See Also:
    * [`pnorm()`][amltk.distances.pnorm]
"""

euclidean_distance = l2_distance
"""Calculates the euclidean distance between each column in x and y.

Same as [`l2_distance()`][amltk.distances.l2_distance].
"""

NamedDistance: TypeAlias = Literal["l1", "l2", "euclidean", "cosine", "max"]
"""Predefined distance metrics.

Possible values are:

* `"l1"`: [`l1_distance()`][amltk.distances.l1_distance]
* `"l2"`: [`l2_distance()`][amltk.distances.l2_distance]
* `"euclidean"`: [`euclidean_distance()`][amltk.distances.euclidean_distance]
* `"cosine"`: [`cosine_distance()`][amltk.distances.cosine_distance]
* `"max"`: [`linf_distance()`][amltk.distances.linf_distance]
"""

distance_metrics: dict[NamedDistance, DistanceMetric] = {
    "l1": l1_distance,
    "l2": l2_distance,
    "euclidean": euclidean_distance,
    "cosine": cosine_distance,
    "max": partial(pnorm, p=np.inf),
}


class NearestNeighborsDistance:
    """Uses [sklearn.neighbors.NearestNeighbors][] to calculate the distance."""

    def __init__(self, **nn_kwargs: Any):
        """Creates a new NearestNeighborsDistance.

        Args:
            **nn_kwargs: Keyword arguments to pass to
                [sklearn.neighbors.NearestNeighbors][].
        """
        super().__init__()
        self.nn_kwargs = nn_kwargs

    def __call__(
        self,
        x: npt.ArrayLike,
        y: npt.ArrayLike,
    ) -> npt.NDArray[np.floating]:
        """Calculates the distance between each column in x and y.

        Args:
            x: An array-like with columns being the features and rows being the samples.
            y: A array with the same index as x.

        Returns:
            An array with the same index as x.
        """
        from sklearn.neighbors import NearestNeighbors

        self.nn = NearestNeighbors(**self.nn_kwargs)

        _x = np.asarray(x)
        _y = np.asarray(y)

        if _y.ndim != 1:
            raise ValueError(f"y must be a 1-dimensional array. Got shape {_y.shape}")

        _y = _y.reshape(1, -1)
        _x = _x.T

        if _x.ndim == 1:
            _x = np.asarray([_x])

        self.nn.fit(_x)
        distances, _ = self.nn.kneighbors(
            _y,
            n_neighbors=len(_x),
            return_distance=True,
        )
        return np.asarray(distances.reshape(-1), dtype=float)



"""The [`NEPSOptimizer`][amltk.optimization.optimizers.neps.NEPSOptimizer],
is a wrapper around the [`NePs`](https://github.com/automl/neps) optimizer.

!!! tip "Requirements"

    This requires `smac` which can be installed with:

    ```bash
    pip install amltk[neps]

    # Or directly
    pip install neural-pipeline-search
    ```

!!! warning "NePs is still in development"

    NePs is still in development and is not yet stable.
    There are likely going to be issues. Please report any issues to NePs or in
    AMLTK.

This uses `ConfigSpace` as its [`search_space()`][amltk.pipeline.Node.search_space] to
optimize.

Users should report results using
[`trial.success(loss=...)`][amltk.optimization.Trial.success]
where `loss=` is a scaler value to **minimize**. Optionally,
you can also return a `cost=` which is used for more budget aware algorithms.
Again, please see NeP's documentation for more.

!!! warning "Conditionals in ConfigSpace"

    NePs does not support conditionals in its search space. This is account
    for when using the
    [`preferred_parser()`][amltk.optimization.optimizers.neps.NEPSOptimizer.preferred_parser].
    during search space creation. In this case, it will simply remove all conditionals
    from the search space, which may not be ideal for the given problem at hand.

Visit their documentation for what you can pass to
[`NEPSOptimizer.create()`][amltk.optimization.optimizers.neps.NEPSOptimizer.create].

The below example shows how you can use neps to optimize an sklearn pipeline.

```python exec="False" source="material-block" result="python"
from __future__ import annotations

import logging

from sklearn.datasets import load_iris
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import accuracy_score
from sklearn.model_selection import train_test_split

from amltk.optimization.optimizers.neps import NEPSOptimizer
from amltk.scheduling import Scheduler
from amltk.optimization import History, Trial
from amltk.pipeline import Component

logging.basicConfig(level=logging.INFO)


def target_function(trial: Trial, pipeline: Pipeline) -> Trial.Report:
    X, y = load_iris(return_X_y=True)
    X_train, X_test, y_train, y_test = train_test_split(X, y)
    clf = pipeline.configure(trial.config).build("sklearn")

    with trial.begin():
        clf.fit(X_train, y_train)
        y_pred = clf.predict(X_test)
        accuracy = accuracy_score(y_test, y_pred)
        loss = 1 - accuracy
        return trial.success(loss=loss, accuracy=accuracy)

    return trial.fail()
from amltk._doc import make_picklable; make_picklable(target_function)  # markdown-exec: hide


pipeline = Component(RandomForestClassifier, space={"n_estimators": (10, 100)})
space = pipeline.search_space(parser=NEPSOptimizer.preferred_parser())
optimizer = NEPSOptimizer.create(space=space)

N_WORKERS = 2
scheduler = Scheduler.with_processes(N_WORKERS)
task = scheduler.task(target_function)

history = History()

@scheduler.on_start(repeat=N_WORKERS)
def on_start():
    trial = optimizer.ask()
    task.submit(trial, pipeline)

@task.on_result
def tell_and_launch_trial(_, report: Trial.Report):
    if scheduler.running():
        optimizer.tell(report)
        trial = optimizer.ask()
        task.submit(trial, pipeline)

@task.on_result
def add_to_history(_, report: Trial.Report):
    history.add(report)

scheduler.run(timeout=3, wait=False)

print(history.df())
```

!!! todo "Deep Learning"

    Write an example demonstrating NEPS with continuations

!!! todo "Graph Search Spaces"

    Write an example demonstrating NEPS with its graph search spaces
"""  # noqa: E501
from __future__ import annotations

import logging
import shutil
from collections.abc import Mapping, Sequence
from copy import deepcopy
from dataclasses import dataclass
from datetime import datetime
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, Any, Protocol
from typing_extensions import override

import metahyper.api
from ConfigSpace import ConfigurationSpace
from metahyper import instance_from_map
from more_itertools import first_true
from neps.optimizers import SearcherMapping
from neps.search_spaces.parameter import Parameter
from neps.search_spaces.search_space import SearchSpace, pipeline_space_from_configspace

import amltk.randomness
from amltk.optimization import Optimizer, Trial
from amltk.pipeline import Node
from amltk.pipeline.parsers.configspace import parser as configspace_parser
from amltk.store import PathBucket

if TYPE_CHECKING:
    from typing_extensions import Self

    from neps.api import BaseOptimizer

    from amltk.optimization.metric import Metric
    from amltk.types import Seed

    class NEPSPreferredParser(Protocol):
        """The preferred parser call signature for NEPSOptimizer."""

        def __call__(
            self,
            node: Node,
            *,
            seed: int | None = None,
            flat: bool = False,
            delim: str = ":",
        ) -> ConfigurationSpace:
            """See [`configspace_parser`][amltk.pipeline.parsers.configspace.parser]."""
            ...


logger = logging.getLogger(__name__)


@dataclass
class NEPSTrialInfo:
    """The info for a trial."""

    name: str
    config: dict[str, Any]
    pipeline_directory: Path
    previous_pipeline_directory: Path | None


def _to_neps_space(
    space: SearchSpace
    | ConfigurationSpace
    | Mapping[str, ConfigurationSpace | Parameter],
) -> SearchSpace:
    if isinstance(space, SearchSpace):
        return space

    try:
        # Support pipeline space as ConfigurationSpace definition
        if isinstance(space, ConfigurationSpace):
            space = pipeline_space_from_configspace(space)

        # Support pipeline space as mix of ConfigurationSpace and neps parameters
        new_pipeline_space: dict[str, Parameter] = {}
        for key, value in space.items():
            if isinstance(value, ConfigurationSpace):
                config_space_parameters = pipeline_space_from_configspace(value)
                new_pipeline_space = {**new_pipeline_space, **config_space_parameters}
            else:
                new_pipeline_space[key] = value
        space = new_pipeline_space

        # Transform to neps internal representation of the pipeline space
        return SearchSpace(**space)
    except TypeError as e:
        message = f"The pipeline_space has invalid type: {type(space)}"
        raise TypeError(message) from e


def _to_neps_searcher(
    *,
    space: SearchSpace,
    searcher: str | BaseOptimizer | None = None,
    loss_value_on_error: float | None = None,
    cost_value_on_error: float | None = None,
    max_cost_total: float | None = None,
    ignore_errors: bool = False,
    searcher_kwargs: Mapping[str, Any] | None = None,
) -> BaseOptimizer:
    if searcher == "default" or searcher is None:
        if space.has_fidelity:
            searcher = "hyperband"
            if hasattr(space, "has_prior") and space.has_prior:
                searcher = "hyperband_custom_default"
        else:
            searcher = "bayesian_optimization"
        logger.info(f"Running {searcher} as the searcher")

    _searcher_kwargs = dict(searcher_kwargs) if searcher_kwargs else {}
    _searcher_kwargs.update(
        {
            "loss_value_on_error": loss_value_on_error,
            "cost_value_on_error": cost_value_on_error,
            "ignore_errors": ignore_errors,
        },
    )
    return instance_from_map(SearcherMapping, searcher, "searcher", as_class=True)(
        pipeline_space=space,
        budget=max_cost_total,  # TODO: use max_cost_total everywhere
        **_searcher_kwargs,
    )


class NEPSOptimizer(Optimizer[NEPSTrialInfo]):
    """An optimizer that uses SMAC to optimize a config space."""

    def __init__(
        self,
        *,
        space: SearchSpace,
        loss_metric: Metric,
        cost_metric: Metric | None = None,
        optimizer: BaseOptimizer,
        working_dir: Path,
        seed: Seed | None = None,
        bucket: PathBucket | None = None,
    ) -> None:
        """Initialize the optimizer.

        Args:
            space: The space to use.
            loss_metric: The metric to optimize.
            cost_metric: The cost metric to use. Only certain NePs optimizers support
            optimizer: The optimizer to use.
            seed: The seed to use for the trials (and not optimizers).
            working_dir: The directory to use for the trials.
            bucket: The bucket to give to trials generated from this optimizer.
        """
        if isinstance(loss_metric, Sequence):
            raise ValueError("NePs does not support multiple metrics")

        if cost_metric is not None and cost_metric.minimize is False:
            raise ValueError("NePs only supports minimizing cost metrics")

        if cost_metric is None and optimizer.budget is not None:
            raise ValueError(
                "NePs optimizers with a budget require a cost metric to be provided",
            )

        metrics = [loss_metric]
        if cost_metric is not None:
            metrics.append(cost_metric)

        super().__init__(bucket=bucket, metrics=metrics)
        self.space = space
        self.seed = amltk.randomness.as_int(seed)
        self.optimizer = optimizer
        self.working_dir = working_dir
        self.loss_metric = loss_metric
        self.cost_metric = cost_metric

        self.optimizer_state_file = self.working_dir / "optimizer_state.yaml"
        self.base_result_directory = self.working_dir / "results"
        self.serializer = metahyper.utils.YamlSerializer(self.optimizer.load_config)

        self.working_dir.mkdir(parents=True, exist_ok=True)
        self.base_result_directory.mkdir(parents=True, exist_ok=True)

    @classmethod
    def create(  # noqa: PLR0913
        cls,
        *,
        space: (
            SearchSpace
            | ConfigurationSpace
            | Mapping[str, ConfigurationSpace | Parameter]
            | Node
        ),
        metrics: Metric,
        cost_metric: Metric | None = None,
        bucket: PathBucket | str | Path | None = None,
        searcher: str | BaseOptimizer = "default",
        working_dir: str | Path = "neps",
        overwrite: bool = True,
        seed: Seed | None = None,
        max_cost_total: float | None = None,
        searcher_kwargs: Mapping[str, Any] | None = None,
    ) -> Self:
        """Create a new NEPS optimizer.

        Args:
            space: The space to use.
            metrics: The metrics to optimize.

                !!! warning

                    NePs does not support multiple metrics. Please only pass a single
                    metric.

            cost_metric: The cost metric to use. Only certain NePs optimizers support
                this.
            seed: The seed to use for the trials.

                !!! warning

                    NePS optimizers do not support an explicit seeding. If you'd
                    like to seed their optimizers, they use the global
                    `torch.manual_seed`, `np.random.seed`, and `random.seed`.
                    This is not considered a good practice and there is not
                    much we can do from AMLTK to help with this.

            bucket: The bucket to give to trials generated by this optimizer.
            searcher: The searcher to use.
            working_dir: The directory to use for the optimization.
            overwrite: Whether to overwrite the working directory if it exists.
            max_cost_total: The maximum cost to use for the optimization.

                !!! warning

                    This only effects the optimization if the searcher utilizes the
                    budget for it's actual suggestion of the next config. If the
                    searcher does not use the budget. This parameter has no effect.

                    The user is still expected to stop `ask()`'ing for configs when
                    they have reached some budget.
            searcher_kwargs: Additional kwargs to pass to the searcher.
        """
        if isinstance(space, Node):
            space = space.search_space(parser=NEPSOptimizer.preferred_parser())

        match bucket:
            case None:
                bucket = PathBucket(
                    f"{cls.__class__.__name__}-{datetime.now().isoformat()}",
                )
            case str() | Path():
                bucket = PathBucket(bucket)
            case bucket:
                bucket = bucket  # noqa: PLW0127

        space = _to_neps_space(space)
        searcher = _to_neps_searcher(
            space=space,
            searcher=searcher,
            max_cost_total=max_cost_total,
            searcher_kwargs=searcher_kwargs,
        )
        working_dir = Path(working_dir)
        if working_dir.exists() and overwrite:
            logger.info(f"Removing existing working directory {working_dir}")
            shutil.rmtree(working_dir)

        return cls(
            space=space,
            bucket=bucket,
            seed=seed,
            loss_metric=metrics,
            cost_metric=cost_metric,
            optimizer=searcher,
            working_dir=working_dir,
        )

    @override
    def ask(self) -> Trial[NEPSTrialInfo]:
        """Ask the optimizer for a new config.

        Returns:
            The trial info for the new config.
        """
        with self.optimizer.using_state(self.optimizer_state_file, self.serializer):
            (
                config_id,
                config,
                pipeline_directory,
                previous_pipeline_directory,
            ) = metahyper.api._sample_config(  # type: ignore
                optimization_dir=self.working_dir,
                sampler=self.optimizer,
                serializer=self.serializer,
                logger=logger,
            )

        match config:
            case SearchSpace():
                _config = config.hp_values()
            case _:  # type: ignore
                _config = {
                    k: v.value if isinstance(v, Parameter) else v
                    for k, v in config.items()  # type: ignore
                }

        info = NEPSTrialInfo(
            name=str(config_id),
            config=deepcopy(_config),
            pipeline_directory=pipeline_directory,
            previous_pipeline_directory=previous_pipeline_directory,
        )

        match self.cost_metric:
            case None:
                metrics = [self.loss_metric]
            case cost_metric:
                metrics = [self.loss_metric, cost_metric]

        trial = Trial(
            name=info.name,
            config=info.config,
            info=info,
            seed=self.seed,
            bucket=self.bucket,
            metrics=metrics,
        )
        logger.debug(f"Asked for trial {trial.name}")
        return trial

    @override
    def tell(self, report: Trial.Report[NEPSTrialInfo]) -> None:
        """Tell the optimizer the result of the sampled config.

        Args:
            report: The report of the trial.
        """
        logger.debug(f"Telling report for trial {report.trial.name}")
        info = report.info
        assert info is not None

        # Get a metric result
        metric_result = first_true(
            report.metric_values,
            pred=lambda value: value.metric.name == self.loss_metric.name,
            default=self.loss_metric.worst,
        )

        # Convert metric result to a minimization loss
        neps_loss: float
        if (_loss := metric_result.distance_to_optimal) is not None:
            neps_loss = _loss
        else:
            neps_loss = metric_result.loss

        result: dict[str, Any] = {"loss": neps_loss}
        metadata: dict[str, Any] = {"time_end": report.time.end}

        if self.cost_metric is not None:
            cost_metric: Metric = self.cost_metric
            _cost = first_true(
                report.metric_values,
                pred=lambda value: value.metric.name == cost_metric.name,
                default=self.cost_metric.worst,
            )
            cost = _cost.value
            result["cost"] = cost

            # If it's a budget aware optimizer
            if self.optimizer.budget is not None:
                with self.optimizer.using_state(
                    self.optimizer_state_file,
                    self.serializer,
                ):
                    self.optimizer.used_budget += cost

                metadata["budget"] = {
                    "max": self.optimizer.budget,
                    "used": self.optimizer.used_budget,
                    "eval_cost": cost,
                    "account_for_cost": True,
                }

        # Dump results
        self.serializer.dump(result, info.pipeline_directory / "result")

        # Load and dump metadata
        config_metadata = self.serializer.load(info.pipeline_directory / "metadata")
        config_metadata.update(metadata)
        self.serializer.dump(config_metadata, info.pipeline_directory / "metadata")

    @override
    @classmethod
    def preferred_parser(cls) -> NEPSPreferredParser:
        """The preferred parser for this optimizer."""
        # TODO: We might want a custom one for neps.SearchSpace, for now we will
        # use config space but without conditions as NePs doesn't support conditionals
        return partial(configspace_parser, conditionals=False)



"""Render a function/class signature."""
from __future__ import annotations

import inspect
from collections.abc import Callable
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Literal
from typing_extensions import override

from amltk._doc import link
from amltk._functional import funcname
from amltk._richutil.renderable import RichRenderable
from amltk.options import _amltk_options

if TYPE_CHECKING:
    from rich.highlighter import Highlighter
    from rich.style import StyleType
    from rich.text import Text


@dataclass
class Function(RichRenderable):
    """Render a function/class signature."""

    f: Callable | type
    signature: bool | Literal["..."] | tuple[tuple, dict] = field(
        default_factory=lambda: _amltk_options.get("rich_signatures", True),
    )
    link: Literal["auto"] | str | Literal[False] = field(
        default_factory=lambda: _amltk_options.get("rich_link", "auto"),
    )
    highlighter: Highlighter | None = None
    prefix: str | None = None
    no_wrap: bool = False

    @override
    def __rich__(self) -> Text:  # noqa: C901, PLR0912, PLR0915
        # Taken from rich._inspect._get_signature
        from rich.highlighter import ReprHighlighter
        from rich.style import Style
        from rich.text import Text

        f = self.f
        qualname = getattr(f, "__qualname__", None)
        if qualname is None:
            qualname = funcname(f, default="<function>")

        is_lambda = False
        if "<lambda>" in qualname:
            qualname = qualname.replace("<lambda>", "λ")
            is_lambda = True

        if self.signature is True:
            try:
                _signature = str(inspect.signature(f))
            except ValueError:
                _signature = "(...)"
        elif self.signature == "...":
            _signature = "(...)"
        elif isinstance(self.signature, tuple):
            _signature = "("
            args, kwargs = self.signature
            if args:
                _signature += ", ".join(map(str, args))
            if kwargs:
                if args:
                    _signature += ", "
                _signature += ", ".join(
                    f"{key}={value}" for key, value in kwargs.items()
                )
            _signature += ")"
        else:
            _signature = "()"

        highlighter = self.highlighter or ReprHighlighter()
        signature_text = highlighter(_signature)

        _name_style: StyleType = "inspect.callable"
        _link = None
        if self.link == "auto":
            _link = link(f)
            if _link is not None:
                _name_style = Style(link=_link, underline=True)
        elif isinstance(self.link, str):
            _link = self.link
            _name_style = Style(link=_link, underline=True)

        # If obj is a module, there may be classes (which are callable) to display
        if self.prefix is not None:
            prefix, style = self.prefix, "inspect.class"
        elif is_lambda:
            prefix, style = "lambda", "yellow italic"
        elif inspect.isclass(f):
            prefix, style = "class", "inspect.class"
        elif inspect.iscoroutinefunction(f):
            prefix, style = "async def", "inspect.async_def"

        else:
            prefix, style = "def", "inspect.def"

        return Text.assemble(
            (f"{prefix} ", style),
            (qualname, _name_style),
            signature_text,
            no_wrap=self.no_wrap,
        )


Class = Function
"""Alias for Function."""



"""Measure things about data."""
from __future__ import annotations

import sys
from collections.abc import Iterable
from typing import Any

import numpy as np
import pandas as pd


def byte_size(data: Any | Iterable[Any]) -> int:
    """Measure the size of data.

    Works for numpy-arrays, pandas DataFrames and Series, and iterables of any of
    these.

    Args:
        data: The data to measure.

    Returns:
        The size of the data.
    """
    if isinstance(data, np.ndarray):
        return data.nbytes
    if isinstance(data, pd.DataFrame):
        return int(data.memory_usage(deep=True).sum())
    if isinstance(data, pd.Series):
        return int(data.memory_usage(deep=True))
    if isinstance(data, str):
        return sys.getsizeof(data)
    if isinstance(data, Iterable):
        return sum(byte_size(d) for d in data)

    return sys.getsizeof(data)



"""Options for the AMTLK package.

In general, these options are not intended for functional differences but
more for any output generated by the package, such as rich or logging.
"""
from __future__ import annotations

from collections.abc import Callable
from typing import Any, Literal, TypedDict, TypeVar, overload

from amltk._doc import sklearn_link_generator


class AMLTKOptions(TypedDict):
    """The options available for AMTLK.

    ```python exec="true" source="material-block" result="python"
    from amltk import options

    print(options)
    ```
    """

    rich_signatures: bool
    """Whether to display full signatures in rich output."""

    rich_link: Literal["auto", False]
    """Whether to use links in rich output."""

    links: dict[str, str | Callable[[str], str]]
    """The links to use in rich output.

    The keys are the names of the packages, and the values are either the
    direct link to use or a callable that takes the fully qualified name
    of the object and returns the link to use.
    """

    root_event_manager_name: str


_amltk_options: AMLTKOptions = {
    "rich_signatures": True,
    "rich_link": "auto",
    "links": {"sklearn": sklearn_link_generator},
    "root_event_manager_name": "amltk.root.eventmanager",
}

T = TypeVar("T")


@overload
def get_option(name: str, default: None = None) -> Any | None:
    ...


@overload
def get_option(name: str, default: T) -> Any | T:
    ...


def get_option(name: str, default: T | None = None) -> Any | T | None:
    """Get an option.

    ```python exec="true" source="material-block" result="python"
    from amltk import options

    print(options.get_option("rich_signatures"))
    ```
    """
    return _amltk_options.get(name, default)



"""Defines custom renderers for objects to be displayed nicer with rich."""
from __future__ import annotations

from typing import TYPE_CHECKING
from typing_extensions import override

if TYPE_CHECKING:
    from sklearn.compose import make_column_selector


class rich_make_column_selector:  # noqa: N801
    """Defines better `repr` for make_column_selector when displayed with rich."""

    def __init__(self, col_selector: make_column_selector) -> None:
        """Initialize."""
        super().__init__()
        self.col_selector = col_selector

    @override
    def __repr__(self) -> str:
        _str = ", ".join(
            f"{k}={v}" for k, v in self.col_selector.__dict__.items() if v is not None
        )
        return f"make_column_selector({_str})"



"""A collection of functional programming tools.

This module contains a collection of tools for functional programming.
"""
from __future__ import annotations

from collections.abc import Callable, Hashable, Iterable, Iterator, Mapping, Sequence
from functools import partial, reduce
from inspect import isclass, signature
from itertools import count
from typing import (
    Any,
    Generic,
    TypeAlias,
    TypeVar,
    Union,
)

T = TypeVar("T")
V = TypeVar("V")
V2 = TypeVar("V2")
K = TypeVar("K", bound=Hashable)
VK = TypeVar("VK", bound=Hashable)
RecMapping: TypeAlias = Mapping[K, Union["RecMapping[K, V]", V]]


def prefix_keys(d: Mapping[str, V], prefix: str) -> dict[str, V]:
    """Prefix the keys of a mapping.

    ```python exec="true" source="material-block" result="python" title="prefix_keys"
    from amltk._functional import prefix_keys

    d = {"a": 1, "b": 2}
    print(prefix_keys(d, "c:"))
    ```
    """
    return {prefix + k: v for k, v in d.items()}


def dict_get_not_none(d: Mapping[K, V], key: K, default: V2) -> V | V2:
    """Get a value from a dictionary, or a default value if it is None.

    ```python exec="true" source="material-block" result="python" title="dict_get_not_none"
    from amltk._functional import dict_get_not_none

    d = {"a": None, "b": 2}
    print(dict_get_not_none(d, "a", 1))  # d.get("a", 1) would return None
    ```
    """  # noqa: E501
    if (item := d.get(key)) is not None:
        return item
    return default


def mapping_select(d: Mapping[str, V], prefix: str) -> dict[str, V]:
    """Select a subset of a mapping.

    ```python exec="true" source="material-block" result="python" title="mapping_select"
    from amltk._functional import mapping_select

    d = {"a:b:c": 1, "a:b:d": 2, "c:elephant": 3}
    print(mapping_select(d, "a:b:"))
    # {"c": 1, "d": 2}
    ```

    Args:
        d: The mapping to select from.
        prefix: The prefix to select.

    Returns:
        The selected subset of the mapping.
    """
    return {k[len(prefix) :]: v for k, v in d.items() if k.startswith(prefix)}


def flatten_dict(d: RecMapping[str, V], *, delim: str | None = None) -> dict[str, V]:
    """Flatten a recursive mapping.

    ```python exec="true" source="material-block" result="python" title="flatten_dict"
    from amltk._functional import flatten_dict

    d = {"a": 1, "b": {"c": 2, "d": 3}}
    print(flatten_dict(d))
    # {"a": 1, "b:c": 2, "b:d": 3}
    ```

    Args:
        d: The recursive mapping to flatten.
        delim: The delimiter to use between keys.

    Returns:
        The flattened mapping.
    """
    delim = delim or ":"

    def _flatten_dict(d: RecMapping[str, V], prefix: str) -> Iterator[tuple[str, V]]:
        for k, v in d.items():
            if isinstance(v, Mapping):
                yield from _flatten_dict(v, prefix + k + delim)
            else:
                yield prefix + k, v

    return dict(_flatten_dict(d, ""))


def reverse_enumerate(
    seq: Sequence[T],
    start: int | None = None,
) -> Iterator[tuple[int, T]]:
    """Reverse enumerate.

    This function is similar to enumerate, but it iterates over the
    sequence in reverse.

    ```python exec="true" source="material-block" result="python" title="reverse_enumerate"
    from amltk._functional import reverse_enumerate

    xs = ["a", "b", "c"]
    for i, x in reverse_enumerate(xs):
        print(i, x)
    ```

    Args:
        seq: The sequence to iterate over.
        start: The starting index.

    Returns:
        An iterator over the sequence.
    """  # noqa: E501
    if start is None:
        start = len(seq) - 1
    yield from zip(count(start, -1), reversed(seq))


def rgetattr(obj: Any, attr: str, *args: Any) -> Any:
    """Recursive version of getattr.

    This function is similar to getattr, but it allows you to get
    attributes using '.' notation.

    ```python exec="true" source="material-block" result="python" title="rgetattr"
    from amltk._functional import rgetattr

    class A:
        x = 1

    class B:
        a = A()

    b = B()
    print(rgetattr(b, "a.x"))
    ```

    See Also:
        * https://stackoverflow.com/a/31174427/5332072

    Args:
        obj: The object to get the attribute from.
        attr: The attribute to get.
        *args: The default value to return if the attribute is not found.

    Returns:
        The attribute.
    """

    def _getattr(obj: Any, attr: str) -> Any:
        return getattr(obj, attr, *args)

    return reduce(_getattr, [obj, *attr.split(".")])


def funcname(
    func: Callable,
    default: str | None = None,
    *,
    with_paren: bool = False,
) -> str:
    """Get the name of a function.

    Args:
        func: The function to get the name of.
        default: The default value to return if the name cannot be
            determined automatically.
        with_paren: Whether to include the parentheses.

    Returns:
        The name of the function.
    """
    from amltk.scheduling.task import Task

    if isinstance(func, Task):
        fname = funcname(func.function)
    elif isinstance(func, partial):
        fname = funcname(func.func)
    elif hasattr(func, "__qualname__"):
        fname = func.__qualname__
    elif hasattr(func, "__class__"):
        fname = func.__class__.__name__
    elif default is not None:
        fname = default
    else:
        fname = str(func)

    return f"{fname}()" if with_paren else fname


def fullname(o: Any) -> str:
    """Get the full name of an object.

    Args:
        o: The object to get the name of.

    Returns:
        The full name of the object.
    """
    klass = o if isclass(o) else o.__class__
    if (module := str(klass.__module__)) != "builtins":
        return module + "." + str(klass.__qualname__)

    return str(klass.__qualname__)  # avoid outputs like 'builtins.str'


def classname(c: Any, default: str | None = None) -> str:
    """Get the classname of an object.

    Args:
        c: The item to get the classname of.
        default: The default value to return if the name cannot be
            determined automatically.

    Returns:
        The name of the class.
    """
    if isinstance(c, type):
        return c.__name__
    if hasattr(c, "__class__"):
        return str(c.__class__.__name__)
    if default is not None:
        return default
    return str(c)


def entity_name(
    thing: Any,
    default: str | None = None,
) -> str:
    """Get the name of a thing.

    Args:
        thing: The thing to get the name of.
        default: The default value to return if the name cannot be
            determined automatically.

    Returns:
        The name of the thing.
    """
    if isinstance(thing, str):
        return thing
    if isinstance(thing, type) or hasattr(thing, "__class__"):
        return classname(thing)
    if callable(thing):
        return funcname(thing)
    if hasattr(thing, "__name__"):
        return str(thing.__name__)
    if default is not None:
        return default
    return str(thing)


def callstring(f: Callable, *args: Any, **kwargs: Any) -> str:
    """Get a string representation of a function call.

    Args:
        f: The function to get the string representation of.
        *args: The positional arguments.
        **kwargs: The keyword arguments.

    Returns:
        The string representation of the function call.
    """
    # Iterate over all args, convert them to str, and join them
    args_str = ""
    if any(args):
        args_str += ", ".join(map(str, args))
    if any(kwargs):
        args_str += ", ".join(f"{k}={v}" for k, v in kwargs.items())

    return f"{funcname(f)}({args_str})"


def compare_accumulate(
    xs: Iterable[T],
    op: Callable[[T, T], bool],
    *,
    ffill: bool = False,
) -> Iterator[T]:
    """Compare and accumulate.

    The most recent item to return True is carried forward and yielded.
    Once another item in the iterator returns True, it is then the value to be
    carried forward.

    ```python exec="true" source="material-block" result="python" title="compare_accumulate"
    from amltk._functional import compare_accumulate

    xs = [5, 4, 6, 2, 1, 8]
    print(list(compare_accumulate(xs, lambda x, y: x > y)))
    # [5, 6, 8]

    print(list(compare_accumulate(xs, lambda x, y: x > y, ffill=True)))
    # [5, 5, 6, 6, 6, 8]
    ```

    Args:
        xs: The iterable to compare and accumulate.
        op: The comparison operator.
        ffill: Whether to forward fill. If this is `True`, any item that returns
            `False` will be replaced with the most recent item that returned `True`.
    """  # noqa: E501
    itr = iter(xs)
    current = next(itr, None)
    if current is None:
        return

    yield current
    for x in itr:
        if op(x, current):
            yield x
            current = x
        elif ffill:
            yield current


def transformations(
    x: T,
    transforms: Iterable[Callable[[T], T]],
) -> Iterator[T]:
    """Apply transformations to an object.

    !!! note "Typing"

        The typing specifies that the domain of the type of `x` stays constant, i.e.
        `x` is always of type `T`. However this is not a strict requirement and you
        can safely ignore the type warnings if so.

    ```python exec="true" source="material-block" result="python" title="transforms"
    from amltk._functional import transformations

    def f(x):
        return x + 1

    def g(x):
        return x * 2

    def h(x):
        return x - 1

    x = 1
    steps = list(transformations(x, [f, g, h]))
    print(steps)
    # [1, 2, 4, 3]
    ```

    Args:
        x: The object to transform.
        transforms: The transformations to apply.

    Returns:
        An iterator over the transformed object.
    """
    yield x

    itr = iter(transforms)
    t = next(itr, None)
    if t is not None:
        yield from transformations(t(x), itr)


def select_by_signature(
    f: Callable,
    kwargs: Mapping[str, Any],
) -> dict[str, Any]:
    """Select the kwargs that are in the signature of `f`.

    Args:
        f: The function to select kwargs for.
        kwargs: The kwargs to select from.

    Returns:
        The selected kwargs.
    """
    if not any(kwargs):
        return {}

    sig = signature(f)
    return {k: v for k, v in kwargs.items() if k in sig.parameters}


class Flag(Generic[T]):
    """A flag.

    This class is used to store a value that can be reset to its
    initial value.

    ```python
    flag = Flag(1)
    flag.value  # 1

    flag.set(2)
    flag.value  # 2

    flag.reset()
    flag.value  # 1
    ```

    Args:
        initial: The initial value.

    Attributes:
        value: The current flag value.
        initial: The initial value.
    """

    def __init__(self, initial: T) -> None:
        """Initialize the flag."""
        super().__init__()
        self.value = initial
        self.initial = initial

    def reset(self) -> None:
        """Reset the flag to its initial value."""
        self.value = self.initial

    def set(self, value: T) -> None:
        """Set the flag value."""
        self.value = value

    def __bool__(self) -> bool:
        """Get the flag value."""
        return bool(self.value)



"""Reduce the dtypes of data."""
from __future__ import annotations

import logging
from typing import TypeAlias, TypeVar

import numpy as np
import pandas as pd

logger = logging.getLogger(__name__)

DataContainer: TypeAlias = np.ndarray | (pd.DataFrame | pd.Series)
D = TypeVar("D", bound=DataContainer)


def reduce_floating_precision(x: D) -> D:
    """Reduce the floating point precision of the data.

    For a float array, will reduce by one step, i.e. float32 -> float16, float64
    -> float32.

    Args:
        x: The data to reduce.

    Returns:
        The reduced data.
    """
    # For a dataframe, we recurse over all columns
    if isinstance(x, pd.DataFrame):
        # Using `apply` doesn't work
        for col in x.columns:
            x[col] = reduce_floating_precision(x[col])
        return x  # type: ignore

    if x.dtype.kind != "f":
        return x

    _reduction_map = {
        # Base numpy dtypes
        "float128": "float64",
        "float96": "float64",
        "float64": "float32",
        "float32": "float16",
        # Nullable pandas dtypes (only supports 64 and 32 bit)
        "Float64": "Float32",
    }

    if (dtype := _reduction_map.get(x.dtype.name)) is not None:
        return x.astype(dtype)  # type: ignore

    return x


def reduce_int_span(x: D) -> D:
    """Reduce the integer span of the data.

    For an int array, will reduce to the smallest dtype that can hold the
    minimum and maximum values of the array.

    Args:
        x: The data to reduce.

    Returns:
        The reduced data.
    """
    # For a dataframe, we recurse over all columns
    if isinstance(x, pd.DataFrame):
        # Using `apply` doesn't work
        for col in x.columns:
            x[col] = reduce_int_span(x[col])
        return x  # type: ignore

    if x.dtype.kind not in "iu":
        return x

    min_dtype = np.min_scalar_type(x.min())  # type: ignore
    max_dtype = np.min_scalar_type(x.max())  # type: ignore
    dtype = np.result_type(min_dtype, max_dtype)
    return x.astype(dtype)  # type: ignore


def reduce_dtypes(x: D, *, reduce_int: bool = True, reduce_float: bool = True) -> D:
    """Reduce the dtypes of data.

    When a dataframe, will reduce the dtypes of all columns.
    When applied to an iterable, will apply to all elements of the iterable.

    For an int array, will reduce to the smallest dtype that can hold the
    minimum and maximum values of the array. Otherwise for floats, will reduce
    by one step, i.e. float32 -> float16, float64 -> float32.

    Args:
        x: The data to reduce.
        reduce_int: Whether to reduce integer dtypes.
        reduce_float: Whether to reduce floating point dtypes.
    """
    if not isinstance(x, pd.DataFrame | pd.Series | np.ndarray):
        raise TypeError(f"Cannot reduce data of type {type(x)}.")

    if isinstance(x, pd.Series | pd.DataFrame):
        x = x.convert_dtypes()

    if reduce_int:
        x = reduce_int_span(x)
    if reduce_float:
        x = reduce_floating_precision(x)

    return x



"""Renderers for known Executors."""
from __future__ import annotations

from concurrent.futures import ProcessPoolExecutor
from dataclasses import dataclass
from typing import TYPE_CHECKING
from typing_extensions import override

from amltk._richutil.renderable import RichRenderable

if TYPE_CHECKING:
    from rich.panel import Panel


@dataclass
class ProcessPoolExecutorRenderer(RichRenderable):
    """Render a ProcessPoolExecutor."""

    executor: ProcessPoolExecutor

    @override
    def __rich__(self) -> Panel:
        from rich.panel import Panel
        from rich.pretty import Pretty
        from rich.text import Text

        max_workers = self.executor._max_workers  # type: ignore
        return Panel.fit(
            Pretty({"max_workers": max_workers}),
            title=Text("ProcessPoolExecutor", no_wrap=True),
            title_align="left",
        )
